"""
# ============================================================================
# Create dynamic menus (dependent on the state of the Schedule object)
#
# NOTE: this module is dependent on functions in EditCourses
#       edit_block_dialog(selected_object, tree_id)
#       add_blocks_dialog(selected_object, tree_id)
#       edit_section_dialog(selected_object, tree_id)
#       edit_course_dialog(selected_object, tree_id)
#       remove_selected_from_parent(parent_object, selected_object, tree_parent_id)
#       modify_course_needs_allocation(selected_object, true_false, tree_id)
#       add_section_dialog(selected_object, tree_id)
#       assign_selected_to_parent(selected_object, resource, tree_id)
#       remove_selected_from_parent(selected_object, resource, tree_id)
#       remove_all_types_from_selected_object(resource_type_string, selected_object, tree_id)
#
# ============================================================================
"""
from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

from ..gui_generics.menu_and_toolbars import MenuItem, MenuType
from ..model import Course, Teacher, ResourceType, Section, Block, Lab, Stream, Schedule

if TYPE_CHECKING:
    from .edit_courses import EditCourses


# ====================================================================================================================
# Creates all the pop menu items, and their sub-menus if required
# ... each menu type will then call the appropriate 'presenter' code that will manage the modifications to the
#     tree/list menus in the EditCourse gui
# ====================================================================================================================
class CreateTreePopupMenuActions:
    """Creates all the pop menu items, dependent on the selected object, and their sub-menus if required

        Each menu type will then call the appropriate 'presenter' code that will manage the modifications to the
        tree/list menus in the EditCourse gui
    """
    def __init__(self, presenter: EditCourses, selected_object, parent_object, tree_id: str, tree_parent_id: str):
        """
        :param presenter: The presenter that is handling all the logic
        :param selected_object: The object that is the target of this menu
        :param parent_object: The parent of the object (ex. which block this teacher is attached to)
        :param tree_id: the id of the selected tree object
        :param tree_parent_id: the id of the selected tree object
        """
        self.presenter = presenter
        self.selected_object = selected_object
        self.parent_object = parent_object if parent_object is not None else presenter.schedule
        self.tree_id = tree_id
        self.tree_parent_id = tree_parent_id

    # ================================================================================================================
    # Create pop up menu based on the selected object
    # ================================================================================================================
    def create_tree_popup_menus(self):
        """create popup menus for trees

        menus are dynamic, depending on the current schedule, and what tree item was selected
        """

        menu_list: list[MenuItem] = []

        # ------------------------------------------------------------------------------------------------------------
        # course
        # ------------------------------------------------------------------------------------------------------------
        if isinstance(self.selected_object, Course):
            self._edit_course(menu_list)
            self._remove_item(menu_list, "Remove Course")

            menu_list.append(MenuItem(menu_type=MenuType.Separator))

            self._needs_allocation(menu_list)

            menu_list.append(MenuItem(menu_type=MenuType.Separator))

            self._add_section( menu_list)
            self._sub_menu_add_items(menu_list, ResourceType.teacher)
            self._sub_menu_add_items(menu_list, ResourceType.lab)
            self._sub_menu_add_items(menu_list, ResourceType.stream)

            menu_list.append(MenuItem(menu_type=MenuType.Separator))

            self._remove_all_items(menu_list, "course")

        # ------------------------------------------------------------------------------------------------------------
        # section
        # ------------------------------------------------------------------------------------------------------------
        if isinstance(self.selected_object, Section):
            self._edit_section(menu_list)
            self._remove_item(menu_list, "Remove Section")

            menu_list.append(MenuItem(menu_type=MenuType.Separator))

            self._add_blocks(menu_list)
            self._sub_menu_add_items(menu_list, ResourceType.teacher)
            self._sub_menu_add_items(menu_list, ResourceType.lab)
            self._sub_menu_add_items(menu_list, ResourceType.stream)

            menu_list.append(MenuItem(menu_type=MenuType.Separator))
            self._sub_menu_remove_items(menu_list, ResourceType.teacher)
            self._sub_menu_remove_items(menu_list, ResourceType.lab)
            self._sub_menu_remove_items(menu_list, ResourceType.stream)
            self._remove_all_items(menu_list, "section")

        # ------------------------------------------------------------------------------------------------------------
        # block
        # ------------------------------------------------------------------------------------------------------------
        if isinstance(self.selected_object, Block):
            self._edit_block(menu_list)
            self._remove_item(menu_list, "Remove Class Time")

            menu_list.append(MenuItem(menu_type=MenuType.Separator))

            self._sub_menu_add_items(menu_list, ResourceType.teacher)
            self._sub_menu_add_items(menu_list, ResourceType.lab)

            menu_list.append(MenuItem(menu_type=MenuType.Separator))
            self._sub_menu_remove_items(menu_list, ResourceType.teacher)
            self._sub_menu_remove_items(menu_list, ResourceType.lab)
            self._remove_all_items(menu_list, "block")

        # ------------------------------------------------------------------------------------------------------------
        # lab/teacher
        # ------------------------------------------------------------------------------------------------------------
        if isinstance(self.selected_object, Teacher):
            self._remove_item(menu_list, "Remove Teacher")
        if isinstance(self.selected_object, Lab):
            self._remove_item(menu_list, "Remove Lab")

        return menu_list

    # ================================================================================================================
    # Private methods that create the MenuList items
    # ================================================================================================================

    def _edit_block(self, menu_list: list[MenuItem]):
        menu_list.append(
            MenuItem(menu_type=MenuType.Command, label="Edit Class Time",
                     command=lambda *_: self.presenter.edit_block_dialog(self.selected_object, self.tree_id, self.tree_parent_id)
                     )
        )

    # ------------------------------------------------------------------------------------------------------------
    def _add_blocks(self, menu_list: list[MenuItem]):
        menu_list.append(
            MenuItem(menu_type=MenuType.Command, label="Add Class Times",
                     command=lambda *_: self.presenter.add_blocks_dialog(self.selected_object, self.tree_id)
                     )

        )

    # ------------------------------------------------------------------------------------------------------------
    def _edit_section(self, menu_list: list[MenuItem]):
        menu_list.append(
            MenuItem(menu_type=MenuType.Command, label="Edit Section",
                     command=lambda *_: self.presenter.edit_section_dialog(self.selected_object, self.tree_id)
                     )
        )

    # ------------------------------------------------------------------------------------------------------------
    def _edit_course(self, menu_list: list[MenuItem]):
        menu_list.append(
            MenuItem(menu_type=MenuType.Command, label="Edit Course",
                     command=lambda *_: self.presenter.edit_course_dialog(self.selected_object, self.tree_id)
                     )
        )

    # ------------------------------------------------------------------------------------------------------------
    def _remove_item( self, menu_list: list[MenuItem], text):
        menu_list.append(
            MenuItem(menu_type=MenuType.Command, label=text,
                     command=lambda *_: self.presenter.remove_selected_from_parent(self.parent_object, self.selected_object, self.tree_parent_id))
        )

    # ------------------------------------------------------------------------------------------------------------
    def _needs_allocation(self, menu_list: list[MenuItem]):
        if self.selected_object.needs_allocation:
            menu_list.append(
                MenuItem(menu_type=MenuType.Command, label="Unset 'needs allocation'",
                         command=lambda *_: self.presenter.modify_course_needs_allocation(self.selected_object, False, self.tree_id))
            )
        else:
            menu_list.append(
                MenuItem(menu_type=MenuType.Command, label="Set 'needs allocation'",
                         command=lambda *_: self.presenter.modify_course_needs_allocation(self.selected_object, True, self.tree_id))
            )

    # ------------------------------------------------------------------------------------------------------------
    def _add_section(self, menu_list: list[MenuItem]):
        menu_list.append(MenuItem(menu_type=MenuType.Command, label="Add Sections",
                     command=lambda *_: self.presenter.add_section_dialog(self.selected_object, self.tree_id))
        )

    # ------------------------------------------------------------------------------------------------------------
    def _sub_menu_add_items(self, menu_list: list[MenuItem], view: ResourceType):
        match view:
            case ResourceType.lab:
                items = [o for o in self.presenter.schedule.labs() if not self.selected_object.has_lab(o)]
                text = "Add Lab"
            case ResourceType.teacher:
                items = [o for o in self.presenter.schedule.teachers() if not self.selected_object.has_teacher(o)]
                text = "Add Teacher"
            case ResourceType.stream:
                items = [o for o in self.presenter.schedule.streams() if not self.selected_object.has_stream(o)]
                text = "Add Stream"
            case _:
                items = []
                text = "Add generic item"

        sub_menu = MenuItem(menu_type=MenuType.Cascade, tear_off=False, label=text)

        # need to define new sub in each iteration, because otherwise the closures will get messed up
        # ... do NOT use lambda's in a loop, we need to bind the individual 'item' in loop rather than its last value
        # ... https://stackoverflow.com/questions/54288926/python-loops-and-closures
        for item in items:
            def f(item_closure= item):
                return  self.presenter.assign_selected_to_parent(self.selected_object, item_closure,
                                                                                               self.tree_id)
            sub_menu.add_child(MenuItem(menu_type=MenuType.Command, label=str(item), command=f))
        menu_list.append(sub_menu)

    # ------------------------------------------------------------------------------------------------------------
    def _sub_menu_remove_items(self, menu_list: list[MenuItem], view: ResourceType):
        match view:
            case ResourceType.lab:
                items = [o for o in self.presenter.schedule.labs() if self.selected_object.has_lab(o)]
                text = "Remove Lab"
            case ResourceType.teacher:
                items = [o for o in self.presenter.schedule.teachers() if self.selected_object.has_teacher(o)]
                text = "Remove Teacher"
            case ResourceType.stream:
                items = [o for o in self.presenter.schedule.streams() if not self.selected_object.has_stream(o)]
                text = "Remove Stream"
            case _:
                items = []
                text = "Remove generic item"

        sub_menu = MenuItem(menu_type=MenuType.Cascade, tear_off=False, label=text)

        # ... do NOT use lambda's in a loop, we need to bind the individual 'item' in loop rather than its last value
        # ... or use partials
        # ... https://stackoverflow.com/questions/54288926/python-loops-and-closures
        for item in items:
            def f(item_closure=item):
                return self.presenter.remove_selected_from_parent(self.selected_object, item_closure,
                                                                self.tree_id)

            sub_menu.add_child(MenuItem(menu_type=MenuType.Command, label=str(item),
                                        command=f))
        menu_list.append(sub_menu)

    # ------------------------------------------------------------------------------------------------------------
    def _remove_all_items(self, menu_list: list[MenuItem], parent_type):
        sub_menu = MenuItem(menu_type=MenuType.Cascade, tear_off=False, label="Remove All")

        # NOTE: do not put the following in a loop, because the closures won't work properly if you do.
        #       If you don't know what a closure is... https://en.wikipedia.org/wiki/Closure_(computer_programming)
        if parent_type == "course":
            sub_menu.add_child(MenuItem(menu_type=MenuType.Command, label="sections",
                                        command=lambda *_: self.presenter.remove_all_types_from_selected_object("section",
                                                                                                           self.selected_object,
                                                                                                           self.tree_id)
                                        )
                           )

        if parent_type == "section":
            sub_menu.add_child(MenuItem(menu_type=MenuType.Command, label="class times",
                                        command=lambda *_: self.presenter.remove_all_types_from_selected_object("block",
                                                                                                           self.selected_object,
                                                                                                           self.tree_id)
                                        )
                           )

        if parent_type == "course" or parent_type == "section":
            sub_menu.add_child(MenuItem(menu_type=MenuType.Command, label="streams",
                                        command=lambda *_: self.presenter.remove_all_types_from_selected_object("stream",
                                                                                                           self.selected_object,
                                                                                                           self.tree_id)
                                        )
                               )
        sub_menu.add_child(MenuItem(menu_type=MenuType.Command, label="labs",
                                    command=lambda *_: self.presenter.remove_all_types_from_selected_object("lab",
                                                                                                       self.selected_object,
                                                                                                       self.tree_id)
                                    )
                           )
        sub_menu.add_child(MenuItem(menu_type=MenuType.Command, label="teachers",
                                    command=lambda *_: self.presenter.remove_all_types_from_selected_object("teacher",
                                                                                                       self.selected_object,
                                                                                                       self.tree_id)
                                    )
                           )
        menu_list.append(sub_menu)


# ====================================================================================================================
# Creates all the pop menu items, and their sub-menus if required
# ... each menu type will then call the appropriate 'presenter' code that will manage the modifications to the
#     tree/list menus in the EditCourse gui
# ====================================================================================================================
class CreateResourcePopupMenuActions:
    """Creates all the pop menu items, dependent on the selected object, and their sub-menus if required

        Each menu type will then call the appropriate 'presenter' code that will manage the modifications to the
        tree/list menus in the EditCourse gui
    """
    def __init__(self, presenter: EditCourses, schedule: Schedule, resource_type:ResourceType, resource: Lab|Stream|Teacher):
        """
        :param presenter: The presenter that is handling all the logic
        :param resource_type
        :param resource: The object that is the target of this menu
        """
        self.presenter = presenter
        self.resource_type = resource_type
        self.resource = resource
        self.schedule = schedule

    # ================================================================================================================
    # Create pop up menu based on the selected object
    # ================================================================================================================
    def create_resource_popup_menus(self):
        """create popup menus for trees

        menus are dynamic, depending on the current schedule, and what tree item was selected
        """


        # ------------------------------------------------------------------------------------------------------------
        # create a menu for entire friggin course schedule (teacher and labs)
        # ------------------------------------------------------------------------------------------------------------
        menu_course = []
        if self.resource_type != ResourceType.stream:

            for course in self.schedule.courses():
                menu_section = [MenuItem(label=course.number, menu_type=MenuType.Command,
                                         command=partial(self._add_all_sections, self.resource, course))]

                for section in course.sections():
                    menu_block = [MenuItem(label=str(section), menu_type=MenuType.Command,
                                           command=partial(self._add_all_blocks, self.resource, course,
                                                           section))]

                    for block in section.blocks():
                        menu_block.append(MenuItem(label=str(block), menu_type=MenuType.Command,
                                                   command=partial(self._add_course_section_block, self.resource, course, section, block )))
                    menu_section.append(MenuItem(label=str(section), menu_type=MenuType.Cascade,
                                                 children=menu_block))
                menu_course.append(MenuItem(label=course.number, menu_type=MenuType.Cascade, children=menu_section))
            menu_list: list[MenuItem] = [MenuItem(label="Add to Course ...", menu_type=MenuType.Cascade, children=menu_course),]
        else:

            for course in self.schedule.courses():
                menu_section = []

                for section in course.sections():
                    menu_section.append(MenuItem(label=str(section), menu_type=MenuType.Command,
                                           command=partial(self._add_all_blocks, self.resource, course,
                                                           section)))
                menu_course.append(MenuItem(label=course.number, menu_type=MenuType.Cascade, children=menu_section))

            menu_list: list[MenuItem] = [
                MenuItem(label="Add to Course ...", menu_type=MenuType.Cascade, children=menu_course), ]

        return menu_list

    # ================================================================================================================
    # Private methods that create the MenuList items
    # ================================================================================================================

    def _add_all_sections(self, resource, course):
        self.presenter.assign_selected_to_parent(selected=resource, parent=course)

    def _add_all_blocks(self, resource, __course, section):
        self.presenter.assign_selected_to_parent(selected=resource, parent=section)

    def _add_course_section_block(self, resource, __course, __section, block):
        self.presenter.assign_selected_to_parent(selected=resource, parent=block)




"""Creates a horizontally scrollable grid with surrounding data frames that are not scrollable"""
from functools import partial
from typing import Optional, Literal, Callable, Any

import tkinter as tk
import tkinter.ttk as ttk
from tkinter.messagebox import showerror

from ..modified_tk import get_fonts_and_colours
from ..modified_tk import Pane

from ..Utilities import Colour
from ..gui_generics.number_validations import entry_float
from ..modified_tk import Hovertip


# =====================================================================================================================
# globals
# =====================================================================================================================

# colours
HEADER_COLOUR1           = "#abcdef"
HEADER_COLOUR2           = Colour.lighten(HEADER_COLOUR1, 10)
VERY_LIGHT_GREY          = "#dddddd"
ROW_COL_INDICATOR_COLOUR = Colour.lighten("#cdefab", 5)
SUMMARY_HEADER_COLOUR    = Colour.string("lemonchiffon")
SUMMARY_COLOUR           = Colour.lighten(SUMMARY_HEADER_COLOUR, 5)
FG_COLOUR                = "black"
BG_COLOUR                = "white"
NOT_OK_COLOUR            = Colour.string("light salmon")
NOT_OK_DARK_COLOUR       = Colour.add(NOT_OK_COLOUR, VERY_LIGHT_GREY)
FRAME_BACKGROUND         = "black"  # provides the borders around the entry widgets

# layout properties
WIDTH = 5
ENTRY_PADDING = 1
TITLE_WIDTH = 12
SUMMARY_WIDTH = WIDTH + 2
PAD_COL_0 = 0
PAD_COL_1 = 0
PAD_COL_2 = 0
PAD_ROW_0 = 0
PAD_ROW_1 = 0
PAD_ROW_2 = 0

# entry widget properties
ENTRY_PROPS = {
    'width': WIDTH,
    'relief': 'ridge',
    'border': 0,
    'justify': 'center',
    'fg': FG_COLOUR,
    'disabledforeground': FG_COLOUR,
    'highlightbackground': BG_COLOUR,
    'bg': BG_COLOUR,
    'disabledbackground': BG_COLOUR
}


# trigger for data change
DATA_CHANGED: dict[tuple[int,int], tk.StringVar] = dict()
def on_var_change(v, row, col, *_):
    DATA_CHANGED[row,col] = v


# =====================================================================================================================
# Allocation Grid Tk
# =====================================================================================================================
class AllocationGridTk:

    @property
    def num_summary_col(self):
        return len(self.summary_header_widgets)

    @property
    def num_summary_sub_col(self):
        return len(self.summary_sub_header_widgets)

    @property
    def num_rows(self):
        return len(self.title_widgets)

    @property
    def num_cols(self):
        return len(self.sub_header_widgets)


    # -----------------------------------------------------------------------------------------------------------------
    # constructor
    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self, frame: tk.Frame, rows,
                 col_merge: list[int],
                 summary_merge: list[int],
                 cb_process_data_change = lambda *_, **__: True,
                 bottom_cell_valid: Callable[[Any], bool] = lambda *_: True
                 ):
        """

        :param frame: where to put this grid
        :param rows: number of rows in the grid
        :param col_merge: list, each item represents a group of columns (affects colouring)
        :param summary_merge: list, each item represents a group of columns in the totals sections
        :param cb_process_data_change:


        Column Merge Example: if you want this for your 2 heading rows
    
        +-------------+----------+--------------------+
        | heading1    | heading2 | heading3           |
        +------+------+----------+------+------+------+
        | sub1 | sub2 | sub1     | sub1 | sub2 | sub3 |
        +------+------+----------+------+------+------+

        use col_merge = [2,1,3]

        """

        # get rid of anything that is currently on this frame
        for w in frame.winfo_children():
            w.destroy()

        # keep this for later
        self.frame = frame
        self.process_data_change_handler = cb_process_data_change
        self.panes = []
        self.bottom_cell_valid = bottom_cell_valid

        # set up the font for the entry widgets
        _, self.fonts = get_fonts_and_colours()

        ENTRY_PROPS['font'] = self.fonts.normal

        # make the frames
        self._layout(frame)

        # make the other stuff
        self.header_widgets: list[tk.Entry] = []
        self.sub_header_widgets: list[tk.Entry] = []
        self.summary_sub_header_widgets: list[tk.Entry] = []

        self.title_widgets: list[tk.Entry] = []
        self.widgets_row_col: dict[tk.Entry, tuple[int,int]] = dict()
        self.entry_widgets: dict[tuple[int,int], tk.Entry] = dict()
        self.summary_widgets: dict[tuple[int,int], tk.Entry] = dict()
        self.summary_header_widgets: list[tk.Entry] = []

        self.bottom_title_widget: Optional[tk.Entry] = None

        self.bottom_widgets: list[tk.Entry] = []
        self.column_colours: dict[int, str] = dict()

        self.make_header_columns(col_merge)
        self.make_summary_header(summary_merge)
        self.make_row_titles(rows)
        self.make_data_grid(rows, col_merge)
        self.make_summary_grid(rows, summary_merge)
        self.make_bottom_header()
        self.make_bottom(col_merge)


    # -----------------------------------------------------------------------------------------------------------------
    # layout
    # -----------------------------------------------------------------------------------------------------------------
    def _layout(self, frame):
        """
        ┌──────────────────────────────────────────────────────────────────┐
        │       ┌──────────────────────────────────────────────────┐┌────┐ │
        │       │               header                             ││  1 │ │
        │       └──────────────────────────────────────────────────┘└────┘ │
        │ ┌────┐┌──────────────────────────────────────────────────┐┌────┐ │
        │ │ t  ││                                                  ││ s  │ │
        │ │ i  ││                                                  ││ u  │ │
        │ │ t  ││                                                  ││ m  │ │
        │ │ l  ││                data                              ││ m  │ │
        │ │ e  ││                                                  ││ a  │ │
        │ │ s  ││                                                  ││ r  │ │
        │ │    ││                                                  ││ y  │ │
        │ │    ││                                                  ││    │ │
        │ └────┘└──────────────────────────────────────────────────┘└────┘ │
        │ ┌────┐┌──────────────────────────────────────────────────┐       │
        │ │  2 ││               bottom                             │       │
        │ └────┘└──────────────────────────────────────────────────┘       │
        │        <────────────────────────────────────────────────>        │
        └──────────────────────────────────────────────────────────────────┘
        1 = summary header
        2 = bottom title
        """
        
        # NOTE: weird shit happens if outer_frame is a 'Pane' (it's a pain... get it?)
        self.outer_frame = tk.Frame(frame, background="white")
        self.outer_frame.pack(expand=1, fill='both')

        # make the frames
        self.header_pane = Pane(self.outer_frame, background=FRAME_BACKGROUND)
        self.header_frame = self.header_pane.frame
        self.panes.append(self.header_pane)
        self.header_pane.canvas.config(background="white")

        summary_header_pane = Pane(self.outer_frame, background=FRAME_BACKGROUND)
        self.summary_header_frame = summary_header_pane.frame
        self.panes.append(summary_header_pane)

        titles_pane = Pane(self.outer_frame, background=FRAME_BACKGROUND)
        self.titles_frame = titles_pane.frame
        self.panes.append(titles_pane)

        data_pane = Pane(self.outer_frame, background=FRAME_BACKGROUND)
        self.data_pane = data_pane
        self.data_frame = data_pane.frame
        self.data_pane.canvas.config(background="white")
        self.panes.append(data_pane)

        summary_pane = Pane(self.outer_frame, background=FRAME_BACKGROUND)
        self.summary_frame = summary_pane.frame
        self.panes.append(summary_pane)

        bottom_title_pane = Pane(self.outer_frame, background=FRAME_BACKGROUND)
        self.bottom_title_frame = bottom_title_pane.frame
        self.panes.append(bottom_title_pane)

        self.bottom_pane = Pane(self.outer_frame, background=FRAME_BACKGROUND)
        self.bottom_frame = self.bottom_pane.frame
        self.panes.append(self.bottom_pane)
        self.bottom_pane.canvas.config(background="white")


        # make the scrollbars
        scrollbar = ttk.Scrollbar(self.outer_frame, orient="horizontal")
        horizontal_scrollable = (self.header_pane, titles_pane, data_pane, self.bottom_pane)

        self.data_pane.horizontal_scrollbar = scrollbar
        self.header_pane.horizontal_scrollbar = scrollbar
        self.bottom_pane.horizontal_scrollbar = scrollbar

        def sync_horizontal_scroll(*args):
            for w in horizontal_scrollable:
                w.xview(*args)

        scrollbar.config(command=sync_horizontal_scroll)
        for f in horizontal_scrollable:
            f.canvas.config(xscrollcommand=scrollbar.set)


        # configure the layout
        self.outer_frame.columnconfigure(0,weight=0)
        self.outer_frame.columnconfigure(1,weight=100)
        self.outer_frame.columnconfigure(2,weight=0)

        self.header_pane.grid(row=0, column=1, sticky='nsew', pady=PAD_ROW_0, padx=PAD_COL_1)
        summary_header_pane.grid(row=0, column=2, sticky='nsew', pady=PAD_ROW_0, padx=PAD_COL_2)
        titles_pane.grid(row=1, column=0, sticky='nsew', pady=PAD_ROW_1, padx=PAD_COL_0)
        data_pane.grid(row=1, column=1, sticky='nsew', pady=PAD_ROW_1, padx=PAD_COL_1)
        summary_pane.grid(row=1, column=2, sticky='nsew', pady=PAD_ROW_1, padx=PAD_COL_2)
        bottom_title_pane.grid(row=2, column=0, sticky='nsew', pady=PAD_ROW_2, padx=PAD_COL_0)
        self.bottom_pane.grid(row=2, column=1, sticky='nsew', pady=PAD_ROW_2,padx=PAD_COL_1)
        scrollbar.grid(row=3, column=1, sticky='nsew', padx=PAD_COL_1)

    # -----------------------------------------------------------------------------------------------------------------
    # make the header columns
    # -----------------------------------------------------------------------------------------------------------------
    def make_header_columns(self, col_merge: list):
        """
        Create disabled entry widgets for the columns (scrollable horizontally).

        Stores widgets in self.header_widgets and self.sub_header_widgets (lists)

        :param col_merge: a list of number of subheadings per heading
        """
        prop = ENTRY_PROPS.copy()
        prop['state'] = 'disabled'

        # merged header
        for i, header in enumerate(col_merge):

            prop['disabledbackground'] = HEADER_COLOUR2 if i % 2 else HEADER_COLOUR1
            prop['highlightbackground'] = HEADER_COLOUR2 if i % 2 else HEADER_COLOUR1

            # frame to hold the merged header, and the subheadings
            mini_frame = tk.Frame(self.header_frame)
            mini_frame.pack(side='left')
            mini_frame.configure(background=self.header_frame.cget("background"))

            # widget
            me = tk.Entry(mini_frame, **prop)
            me.pack(side='top', expand=0, fill='both', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

            # keep these widgets so that they can be configured later
            self.header_widgets.append(me)

            # subsections
            for sub_section in range(header):
                # frame within the mini-frame, so we can stack 'left'
                (hf2 := tk.Frame(mini_frame)).pack(side='left')
                hf2.configure(background=self.header_frame.cget("background"))

                # widget
                (se := tk.Entry(hf2, **prop)).pack(side='left', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

                # keep these widgets so that they can be configured later
                self.sub_header_widgets.append(se)

    # -----------------------------------------------------------------------------------------------------------------
    # summary header
    # -----------------------------------------------------------------------------------------------------------------
    def make_summary_header(self, summary_merge):
        """
        Create disabled entry widgets for summary header

        Stores widgets in self.summary_header_widgets and self.summary_sub_header_widgets (lists)

        :param summary_merge: a list of number of subheadings per heading
        """
        prop = ENTRY_PROPS.copy()
        prop['width'] = SUMMARY_WIDTH
        prop['disabledbackground'] = SUMMARY_HEADER_COLOUR
        prop['highlightbackground'] = SUMMARY_HEADER_COLOUR
        prop['state'] = 'disabled'

        for header in summary_merge:

            # frame to hold the totals header, and the subheadings
            (mini_frame := tk.Frame(self.summary_header_frame)).pack(side='left')
            mini_frame.configure(background=self.header_frame.cget("background"))

            # widget
            (me := tk.Entry(mini_frame, **prop)).pack(side='top', expand=0, fill='both', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

            # keep these widgets so that they can be configured later
            self.summary_header_widgets.append(me)

            for sub_section in range(header):
                # frame within the mini-frame so we can stack left
                (hf2 := tk.Frame(mini_frame)).pack(side='left')
                hf2.configure(background=self.header_frame.cget("background"))

                # widget
                (se := tk.Entry(hf2, **prop)).pack(side='left', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

                # keep these widgets so that they can be configured later
                self.summary_sub_header_widgets.append(se)

    # -----------------------------------------------------------------------------------------------------------------
    # row titles
    # -----------------------------------------------------------------------------------------------------------------
    def make_row_titles(self, rows):
        """
        For each row, create a disabled entry widget to describe each row

        Stores widgets in self.title_widgets (list)

         :param rows: number of rows
        """
        prop = ENTRY_PROPS.copy()
        prop['width'] = TITLE_WIDTH
        prop['justify'] = 'left'
        for _ in range(rows):
            re = tk.Entry(self.titles_frame, **prop, state='disabled', )
            re.pack(side='top', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

            self.title_widgets.append(re)

    # -----------------------------------------------------------------------------------------------------------------
    # data grid
    # -----------------------------------------------------------------------------------------------------------------
    def make_data_grid(self, rows, col_merge: list):
        """
        For each row and column, create an entry widget to hold data

        Stores widgets in self.entry_widgets                (dict[tuple[row,col], tk.Entry])
        Conversely stores row, col in self.widget_row_col   (dict[tk.Entry], tuple[row,col])

        :param rows: number of rows
        :param col_merge: a list of number of subheadings per heading
        """
        prop = ENTRY_PROPS.copy()

        for row in range(rows):
            (df1 := tk.Frame(self.data_frame)).pack(side='top', expand=1, fill='x')
            df1.configure(background=self.data_frame.cget("background"))

            # foreach header
            col: int = 0
            for column_index, header in enumerate(col_merge):
                prop['background'] = VERY_LIGHT_GREY if column_index%2 == 0 else BG_COLOUR
                prop['highlightbackground'] = VERY_LIGHT_GREY if column_index%2 == 0 else BG_COLOUR

                # subsections
                for subsection in range(header):
                    self.column_colours[col] = prop['highlightbackground']

                    # data entry box
                    de = entry_float(df1, textvariable=tk.StringVar(value=""), **prop)
                    de.pack(side='left', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

                    # save row/column with data entry, and vice versa
                    self.entry_widgets[row,col] = de
                    self.widgets_row_col[de] = row, col

                    # binding
                    de.bind("<Tab>", partial(self._move, 'nextCell'))
                    de.bind("<Key-Return>", partial(self._move, 'nextRow'))
                    de.bind("<Shift-Tab>", partial(self._move, 'prevCell'))
                    de.bind("<Key-Up>", partial(self._move, 'prevRow'))
                    de.bind("<Key-uparrow>", partial(self._move, 'prevRow'))
                    de.bind("<Key-Down>", partial(self._move, 'nextRow'))
                    de.bind("<Key-downarrow>", partial(self._move, 'nextRow'))

                    de.bind("<FocusIn>", partial(self.focus_changed, 'focusIn', colour=ROW_COL_INDICATOR_COLOUR))
                    de.bind("<Leave>", self.process_data_change)
                    de.bind("<FocusOut>", partial(self.focus_changed, 'focusOut'))
                    de.bindtags([*de.bindtags(), 1, 0, 2, 3])

                    col += 1

    # -----------------------------------------------------------------------------------------------------------------
    # summary
    # -----------------------------------------------------------------------------------------------------------------
    def make_summary_grid(self, rows, summary_merge):
        """
        create a list of disabled entry widgets to hold summary data per row

        Stores widgets in self.summary_widget (dict[tuple[row,col], tk.Entry]

        :param rows: number of rows
        :param summary_merge: a list of number of subheadings per heading
        """
        prop = ENTRY_PROPS.copy()
        prop['disabledbackground'] = SUMMARY_COLOUR
        prop['highlightbackground'] = SUMMARY_COLOUR
        prop['width'] = SUMMARY_WIDTH
        prop['state'] = 'disabled'

        for row in range(rows):
            (df1 := tk.Frame(self.summary_frame)).pack(side='top', expand=1, fill='x')
            df1.configure(background=self.header_frame.cget("background"))

            # foreach header
            col = 0
            for header in range(len(summary_merge)):

                # subsections
                for _ in range( summary_merge[header]):

                    # data entry box
                    (de := tk.Entry(df1, **prop)).pack(side='left', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

                    # save row/column with totals entry
                    self.summary_widgets[row,col] = de
                    col += 1

    # -----------------------------------------------------------------------------------------------------------------
    # bottom row header
    # -----------------------------------------------------------------------------------------------------------------
    def make_bottom_header(self):
        """
        Create ONE widget to store the title for the bottom row
        """
        prop = ENTRY_PROPS.copy()
        prop['width'] = TITLE_WIDTH

        # widget
        se = tk.Entry(self.bottom_title_frame, **prop, state='disabled')
        se.pack(side='left', expand=1,fill='both', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

        # keep these widgets so that they can be configured later
        self.bottom_title_widget = se


    # -----------------------------------------------------------------------------------------------------------------
    # bottom
    # -----------------------------------------------------------------------------------------------------------------
    def make_bottom(self, col_merge):
        """
        create a list of disabled widgets to hold the summary of the columns

        Stores widgets in self.bottom_widgets (list)

        :param col_merge: a list of number of subheadings per heading
        """

        # merged header
        prop = ENTRY_PROPS.copy()
        col = 0
        for column_index, header in enumerate(col_merge):
            for sub_section in range(header):

                # widget
                prop['disabledbackground'] = HEADER_COLOUR2 if column_index % 2 == 0 else HEADER_COLOUR1
                prop['highlightbackground'] = HEADER_COLOUR2 if column_index % 2 == 0 else HEADER_COLOUR1
                # change Colour every second merged header

                se = tk.Entry(self.bottom_frame, **prop, state='disabled', validate='key')
                se.pack(side='left', padx=ENTRY_PADDING,pady=ENTRY_PADDING)

                # keep these widgets so that they can be configured later
                self.bottom_widgets.append(se)
            col+=1

    # -----------------------------------------------------------------------------------------------------------------
    # populate: assign text variables to each of the entry widgets
    # -----------------------------------------------------------------------------------------------------------------
    def populate(self, header_text: list[str], balloon_text: list[str],
                 sub_header_text: list[str], title_text: list[str],
                 data_vars: dict[tuple[int,int],float], summary_header_texts: list[str],
                 summary_sub_texts: list[str], summary_vars: list[list[str]],
                 bottom_header_text: str, bottom_row_vars: list[str]):
        """
        Enter data into the entry widgets
        :param header_text: A list of major headings in the header frame
        :param balloon_text: A list of text giving further information about the heading
        :param sub_header_text: Sub headings of the header frame
        :param title_text: a list of text describing the title of the row
        :param data_vars: a dictionary where the key is the row,col and the value is the data
        :param summary_header_texts: a list of texts for the summary frame
        :param summary_sub_texts: a list of sub-texts for the summary frame
        :param summary_vars: a 2-d list of variables for the summary data
        :param bottom_header_text: a single string for a title defining the bottom row
        :param bottom_row_vars: a list of data for the bottom row
        :return:
        """
        balloon_text = list(balloon_text)

        # add the tool tips to the header
        for w,text in zip(self.header_widgets, balloon_text):
            if text != "":
                Hovertip(w,text=text)

        # bottom row
        self.bottom_title_widget.configure(textvariable=tk.StringVar(value=bottom_header_text))

        for c in range(len(self.bottom_widgets)):
            self.update_data('bottom',row=0, col=c, value=bottom_row_vars[c])

        # the summary header
        for col in range(self.num_summary_col):
            self.summary_header_widgets[col].configure(textvariable=tk.StringVar(value=summary_header_texts[col]))

        # the summary sub header
        for col in range(self.num_summary_sub_col):
            self.summary_sub_header_widgets[col].configure(textvariable=tk.StringVar(value=summary_sub_texts[col]))

            # the summary data
            for row in range(self.num_rows):
                self.update_data('summary', row, col, summary_vars[row][col])

        # the data grid
        for col in range(self.num_cols):
            for row in range(self.num_rows):
                self.update_data('data', row, col, data_vars[row,col])


        # the header data
        i = 0
        for ht in header_text:
            self.header_widgets[i].configure(textvariable=tk.StringVar(value=ht))
            Hovertip(self.header_widgets[i], balloon_text[i])
            i += 1

        # the sub header data
        i = 0
        for sht in sub_header_text:
            if len(self.sub_header_widgets) > i:
                self.sub_header_widgets[i].configure(textvariable=tk.StringVar(value=sht))
                i += 1
            else:
                break

        # the row titles
        i = 0
        for rht in title_text:
            self.title_widgets[i].configure(textvariable=tk.StringVar(value=rht))
            i += 1

        # get the panes to match their internal structure
        for pane in self.panes:
            pane.after(10, lambda *_: pane.configure_interior(*_))

    # -----------------------------------------------------------------------------------------------------------------
    # navigation routines
    # -----------------------------------------------------------------------------------------------------------------
    def _move(self, where, ev):
        w: tk.Entry = ev.widget

        w.selection_clear()
        row, col = self.widgets_row_col.get(w, (0,0))

        if where == 'nextRow':
            row = min(max(row+1, 0), self.num_rows - 1)
        elif where == 'prevRow':
            row = min(max(row-1, 0), self.num_rows - 1)
        elif where == 'nextCell':
            col = min(max(col+1, 0), self.num_cols - 1)
        elif where == 'prevCell':
            col = min(max(col-1, 0), self.num_cols - 1)

        e = self.entry_widgets.get((row,col), None)
        if e is not None:
            self.set_focus(e)

        # forces Tk to not continue applying any binding routines after this
        return "break"

    # -----------------------------------------------------------------------------------------------------------------
    # what to do when the widget gets the focus
    # -----------------------------------------------------------------------------------------------------------------
    def set_focus(self, e):
        dx, dy = self.data_pane.see(e)
        self.bottom_pane.xview_moveto(dx)
        self.header_pane.xview_moveto(dx)

        e.focus()

    # -----------------------------------------------------------------------------------------------------------------
    # focus has changed - indicate what row/col we are on, callback process change
    # -----------------------------------------------------------------------------------------------------------------
    def focus_changed(self, inout, e, *_, colour = None):
        w: tk.Entry = e.widget
        if inout == "focusIn":
            w.selection_range(0, 'end')
        else:
            w.selection_clear()

        # set Colour and totals Colour
        dcolour = colour or BG_COLOUR
        tcolour = colour

        # get the widget
        r, c = self.widgets_row_col.get(w, (0,0))

        # set colors for rows (data)
        col_colour = Colour.add(dcolour, self.column_colours.get(c))
        for row in range(self.num_rows):
            wid = self.entry_widgets.get((row,c), None)
            if wid is not None:
                wid.configure(bg=col_colour)
                wid.configure(highlightbackground=col_colour)

        # set colors for cols (data)
        for col in range(self.num_cols):
            wid = self.entry_widgets.get((r,col), None)
            if wid is not None:
                col_colour = Colour.add(dcolour, self.column_colours.get(col))
                wid.configure(bg=col_colour)
                wid.configure(highlightbackground=col_colour)

        # set Colour for row header
        widget = self.title_widgets[r]
        widget.configure(disabledbackground=dcolour)
        widget.configure(highlightbackground=dcolour)

        # set colors for summary row
        if tcolour:
            tcolour = Colour.add(tcolour, SUMMARY_COLOUR)
        else:
            tcolour = SUMMARY_COLOUR

        for col in range(self.num_summary_sub_col):
            widget = self.summary_widgets[r,col]
            widget.configure(disabledbackground=tcolour)
            widget.configure(highlightbackground=dcolour)

        self.process_data_change()

    # -----------------------------------------------------------------------------------------------------------------
    # process a data change
    # -----------------------------------------------------------------------------------------------------------------
    def process_data_change(self, *_):
        if len(DATA_CHANGED) == 0:
            return

        for loc, v in DATA_CHANGED.items():
            r,c = loc
            w = self.entry_widgets.get((r, c), None)
            v_str = v.get().strip()
            try:
                if v_str == "":
                    self.process_data_change_handler(r,c,0)
                else:
                    value = float(v_str)
                    self.process_data_change_handler(r,c,value)
                    if value == 0:
                        v.set(value="")
                if w:
                    w.config(foreground=FG_COLOUR)

            except ValueError:
                print ("oh no", v_str)
                if w:
                    w.winfo_toplevel().bell()
                    w.after(100, lambda *args: w.winfo_toplevel().bell(), ())
                    w.after(200, lambda *_: w.winfo_toplevel().bell(), ())
                    w.after(300, lambda *_: w.winfo_toplevel().bell(), ())
                    w.config(foreground="red")
                    showerror("Bad input: ", message="You need to enter a valid float! ", detail=f" '{v_str}' is not a float",
                              icon='error')
                    v.set("")
                    w.focus_set()
                return

        DATA_CHANGED.clear()

    # -----------------------------------------------------------------------------------------------------------------
    # process a data update (can only update widgets in the data/summary/bottom panes
    # -----------------------------------------------------------------------------------------------------------------
    def update_data(self, which: Literal['data','summary','bottom'], row, col, value):
        if which == 'bottom':
            bw = self.bottom_widgets[col]
            bw.configure(textvariable=tk.StringVar(value=value))
            if self.bottom_cell_valid(value):
                bw.configure(highlightbackground=HEADER_COLOUR2)
                bw.configure(disabledbackground = HEADER_COLOUR2)
                if self.column_colours[col] == VERY_LIGHT_GREY:
                    bw.configure(highlightbackground=HEADER_COLOUR1)
                    bw.configure(disabledbackground=HEADER_COLOUR1)
            else:
                bw.configure(highlightbackground=NOT_OK_COLOUR)
                bw.configure(disabledbackground=NOT_OK_COLOUR)
                if self.column_colours[col] == VERY_LIGHT_GREY:
                    bw.configure(highlightbackground=NOT_OK_DARK_COLOUR)
                    bw.configure(disabledbackground=NOT_OK_DARK_COLOUR)

        elif which == 'data':
                widget = self.entry_widgets.get((row, col), None)
                if widget is not None:
                    value = value if float(value) != 0.0 else ""
                    var = tk.StringVar(value=value)
                    widget.configure(textvariable=var)
                    var.trace_add('write', partial(on_var_change, var, row, col))

        elif which == 'summary':
            widget = self.summary_widgets.get((row, col), None)
            if widget:
                widget.configure(textvariable=tk.Variable(value=value))




from __future__ import annotations
import re
from typing import TYPE_CHECKING, Optional
from ..Utilities.id_generator import IdGenerator
from . import WeekDay
from .block import Block, DEFAULT_DURATION, DEFAULT_START, DEFAULT_DAY
from .model_exceptions import InvalidHoursForSectionError

# stuff that we need just for type checking, not for actual functionality
if TYPE_CHECKING:
    from .teacher import Teacher
    from .lab import Lab
    from .stream import Stream
    from .course import Course

DEFAULT_HOURS = 3

section_ids = IdGenerator()


def _validate_hours(hours: float) -> float:
    if hours <= 0:
        raise InvalidHoursForSectionError(f"{hours}: hours must be larger than 0")
    return hours


# ============================================================================
# Section
# ============================================================================
class Section:
    """
    Describes a section (part of a course)
    """
    section_ids = IdGenerator()

    # -------------------------------------------------------------------------
    # CONSTRUCTOR
    # -------------------------------------------------------------------------
    def __init__(self, course: Course, number: str = "", name: str = "",
                 section_id: Optional[int] = None):
        """
        Creates an instance of the Section class.
        :param number: The section's number.
        :param name:
        """

        self._streams: set[Stream] = set()
        self._allocation: dict[Teacher:float] = dict()
        self._blocks: set[Block] = set()

        self.name = name
        self.number = number
        self.num_students: int = 0
        self.course = course

        self._section_id = Section.section_ids.get_new_id(section_id)

    # -------------------------------------------------------------------------
    # PROPERTIES
    # -------------------------------------------------------------------------
    @property
    def hours(self) -> float:
        """
        How many hours does the course have or how many hours does the blocks add up to?
        """
        if self.blocks():
            return sum((b.duration for b in self.blocks()))
        return self.course.hours_per_week

    @property
    def id(self) -> int:
        """ Gets the section's ID """
        return self._section_id

    @property
    def title(self) -> str:
        """ Gets name if defined, otherwise 'Section num' """
        return self.name if self.name else f"Section {self.number}"

    # -------------------------------------------------------------------------
    # conflicts
    # -------------------------------------------------------------------------
    def is_conflicted(self) -> bool:
        """ Checks if there is a conflict with this section """
        return any(map(lambda block: block.conflict.is_conflicted(), self.blocks()))

    # -------------------------------------------------------------------------
    # blocks
    # -------------------------------------------------------------------------
    def blocks(self) -> tuple[Block, ...]:
        """ Gets list of section's blocks """
        return tuple(sorted(self._blocks))

    def remove_block(self, block: Block):
        """ Remove a block from this section """
        self._blocks.discard(block)

    def remove_all_blocks(self):
        for block in self.blocks():
            self.remove_block(block)

    def add_block(self, day: WeekDay | float = DEFAULT_DAY, start: float = DEFAULT_START,
                  duration: float = DEFAULT_DURATION, movable=True, block_id=None) -> Block:
        """ Creates and Assign a block to this section"""
        block = Block(self, day, start, duration, movable=movable, block_id=block_id)
        self._blocks.add(block)
        return block

    def get_block_by_id(self, block_id: int) -> Optional[Block]:
        """ Gets block with given ID from this section """
        blocks = [b for b in self._blocks if b.id == block_id]
        if len(blocks) > 0:
            return blocks[0]
        return None

    def has_block(self, block: Block) -> bool:
        """Does this section have this block? """
        return block in self._blocks

    # -------------------------------------------------------------------------
    # labs
    # -------------------------------------------------------------------------
    def labs(self) -> tuple[Lab, ...]:
        """ Gets all labs assigned to all blocks in this section """
        labs: set[Lab] = set()
        for b in self.blocks():
            labs.update(b.labs())
        return tuple(sorted(labs))

    def add_lab(self, lab: Lab):
        """ Assign a lab to every block in the section """
        for b in self.blocks():
            b.add_lab(lab)
        return self

    def has_lab(self, lab: Lab):
        """Is this lab attached to any of the labs in the blocks? """
        return lab in self.labs()

    def remove_lab(self, lab: Lab):
        """ Remove a lab from every block in the section """
        for b in self.blocks():
            b.remove_lab(lab)

    def remove_all_labs(self):
        """Remove all labs from every block in the section"""
        for lab in self.labs():
            for b in self.blocks():
                b.remove_lab(lab)

    # -------------------------------------------------------------------------
    # teachers (exclusively for scheduler program)
    # -------------------------------------------------------------------------
    def teachers(self) -> tuple[Teacher, ...]:
        """ Gets all teachers assigned to all blocks in this section and
        any teachers that were explicitly assigned to this section"""
        teachers = set()
        for b in self.blocks():
            teachers.update(b.teachers())
        return tuple(sorted(teachers))

    def add_teacher(self, teacher: Teacher):
        """ Assign a teacher to the section """
        for b in self.blocks():
            b.add_teacher(teacher)

    def remove_teacher(self, teacher: Teacher):
        """ Removes teacher from all blocks in this section """
        for b in self.blocks():
            b.remove_teacher(teacher)

    def remove_all_teachers(self):
        """ Removes all teachers from all blocks in this section """
        for t in self.teachers():
            self.remove_teacher(t)

    def has_teacher(self, teacher: Teacher) -> bool:
        """ Checks if section has teacher """
        answer = False
        for b in self.blocks():
            answer = answer or b.has_teacher(teacher)
        return answer

    # -------------------------------------------------------------------------
    # teacher allocation (used for allocation program)
    # -------------------------------------------------------------------------
    def section_defined_teachers(self) -> tuple[Teacher, ...]:
        """gets only teachers that were explicitly assigned to this section in allocation manager"""
        teachers = set()
        teachers.update(self._allocation.keys())
        return tuple(sorted(teachers))

    def remove_allocation(self, teacher):
        self.remove_teacher(teacher)
        if teacher in self._allocation.keys():
            _ = self._allocation.pop(teacher)

    def set_teacher_allocation(self, teacher: Teacher, hours: float):
        """Assign number of hours to teacher for this section. Set hours to 0 to remove
        teacher from this section

        As best as can be done, add teacher to blocks to match the allocation,
        ... if it cannot be done, then set the allocation hours to the given hours
        """
        # clear before resetting
        self.remove_allocation(teacher)

        # don't do anything more if set to zero
        if hours == 0:
            return

        # add to all blocks
        if hours == self.hours and len(self.blocks()) != 0:
            self.add_teacher(teacher)
            return

        # if number of hours is not equal to the total hours of the section,
        # try to find blocks to assign the teacher to
        possible_paths = []
        blocks = self.blocks()
        self._find_block_fit_for_allocation(hours, blocks, "", possible_paths)

        if len(possible_paths) != 0:
            flag = False
            for pp in possible_paths:
                if all(len(b.teachers()) == 0 for tf, b in zip(pp, blocks) if tf == "T"):
                    for block in (b for tf, b in zip(pp, blocks) if tf == "T"):
                        block.add_teacher(teacher)
                    flag = True
                    break
            if not flag:
                for block in (b for tf, b in zip(possible_paths[0], blocks) if tf == "T"):
                    block.add_teacher(teacher)
        else:
            self._allocation[teacher] = hours

    def _find_block_fit_for_allocation(self, hours, blocks, path="", possible_paths=None):
        possible_paths = [] if possible_paths is None else possible_paths
        if hours == 0:
            possible_paths.append(path)
            return
        if len(blocks) == 0:
            return

        block, *blocks = blocks
        if hours >= block.duration:
            new_path = path + "T"
            self._find_block_fit_for_allocation(hours - block.duration, blocks, new_path, possible_paths)

        if hours <= sum((b.duration for b in blocks)):
            new_path = path + "F"
            self._find_block_fit_for_allocation(hours, blocks, new_path, possible_paths)

    def has_allocated_teacher(self, teacher):
        return self.has_teacher(teacher) or teacher in self._allocation.keys()

    def get_teacher_allocation(self, teacher: Teacher) -> float:
        """ Get the number of hours assigned to this teacher for this section """

        # teacher not teaching this section
        if not self.has_teacher(teacher) and teacher not in self._allocation.keys():
            return 0

        # allocation defined
        allocation = self._allocation.get(teacher, 0)

        if len(self.blocks()) != 0:
            allocation += sum((b.duration for b in self.blocks() if teacher in b.teachers()))
        return allocation

    # -------------------------------------------------------------------------
    # streams
    # -------------------------------------------------------------------------
    def streams(self) -> tuple[Stream, ...]:
        """ Gets all streams in this section """
        return tuple(sorted(self._streams))

    def add_stream(self, stream: Stream):
        """ Assign streams to this section. """
        self._streams.add(stream)

    def remove_stream(self, stream: Stream):
        """ Remove stream from this section. """
        self._streams.discard(stream)

    def has_stream(self, stream: Stream) -> bool:
        """Check if a section has a stream """
        return stream in self._streams

    def remove_all_streams(self):
        """ Removes all streams from this section """
        self._streams.clear()

    # -------------------------------------------------------------------------
    # clear everything from the stream
    # -------------------------------------------------------------------------
    def clear(self):
        self.remove_all_teachers()
        self.remove_all_streams()
        self.remove_all_labs()
        self.remove_all_blocks()

    # -------------------------------------------------------------------------
    # string representation
    # -------------------------------------------------------------------------
    def __str__(self) -> str:
        """ Returns a text string that describes the section """
        if self.name and not re.match(r"Section\s*\d*$", self.name):
            return f"Section {self.number}: {self.name}"
        else:
            return f"Section {self.number}"

    def __repr__(self):
        return str(self)

    def __lt__(self, other):
        try:
            a = int(self.number) < int(other.number)
            return a
        except ValueError:
            return self.number < other.number

    def __hash__(self):
        return hash(self._section_id)

    def __eq__(self, other):
        return self.id == other.number



# https://www.pythontutorial.net/tkinter/tkinter-treeview/
import os
import sys
import tkinter as tk
from tkinter import ttk
import csv
bin_dir: str = os.path.dirname(os.path.realpath(__file__))
print (bin_dir)
sys.path.append(os.path.join(bin_dir, "../../"))

from schedule.Tk.Pane import Pane

# TODO: lost the horizontal scrollbar
# TODO: the vertical frames are not the same size (header vs data vs summary)
"""
┌─────────────────────────────────────────────────────────────────────┐
│       ┌──────────────────────────────────────────────────┐          │
│       │               header                             │          │
│       └──────────────────────────────────────────────────┘          │
│ ┌────┐┌──────────────────────────────────────────────────┐┌────┐    │
│ │ t  ││                                                  ││ s  │ ^  │
│ │ i  ││                                                  ││ u  │ |  │
│ │ t  ││                                                  ││ m  │ |  │
│ │ l  ││                data                              ││ m  │ |  │
│ │ e  ││                                                  ││ a  │ |  │
│ │ s  ││                                                  ││ r  │ |  │
│ │    ││                                                  ││ r  │ |  │
│ │    ││                                                  ││ y  │ v  │
│ └────┘└──────────────────────────────────────────────────┘└────┘    │
│       ┌──────────────────────────────────────────────────┐          │
│       │               footer                             │          │
│       └──────────────────────────────────────────────────┘          │
│        <────────────────────────────────────────────────>           │
└─────────────────────────────────────────────────────────────────────┘

horizontal scroll: header data footer
vertical scroll: titles data summary
"""
class AllocationLayout:
    def __init__(self, frame):

        # define the grid for managing this layout
        outer_frame = tk.Frame(frame, background="yellow", borderwidth=5, relief='raised')
        outer_frame.grid_columnconfigure(0, weight=1, pad=5, minsize=20)
        outer_frame.grid_columnconfigure(1, weight=100, pad=5)
        outer_frame.grid_columnconfigure(2, weight=1, pad=5, minsize=20)
        outer_frame.grid_columnconfigure(3, weight=1, pad=5)
        outer_frame.grid_rowconfigure(0, weight=1, pad=5)
        outer_frame.grid_rowconfigure(1, weight=100, pad=5)
        outer_frame.grid_rowconfigure(2, weight=1, pad=5)
        outer_frame.grid_rowconfigure(3, weight=1, pad=5)

        # Define the frames
        self.header_frame = self.scrollable_frame(outer_frame,0,1, height=20, width=None)
        self.title_frame = self.scrollable_frame(outer_frame,1,0)
        self.data_frame = self.scrollable_frame(outer_frame,1,1)
        self.summary_frame = self.scrollable_frame(outer_frame,1,2)
        self.footer_frame = self.scrollable_frame(outer_frame,2,1)

        # Define the scrollbars
        h_scrollbar = tk.Scrollbar(outer_frame, orient='horizontal')
        v_scrollbar = tk.Scrollbar(outer_frame, orient='vertical')
        h_scrollbar.grid(row=3,column=1, sticky='ew')
        v_scrollbar.grid(row=1,column=3, sticky='ns')

        # Link the scrollbars to the frames
        horizontal_scrollable:list[Pane] = [self.footer_frame, self.header_frame, self.data_frame]
        vertical_scrollable:list[Pane] = [self.title_frame, self.data_frame, self.summary_frame]

        def sync_vertical_scroll(*args):
            for w in vertical_scrollable:
                w.yview(*args)

        def sync_horizontal_scroll(*args):
            for w in horizontal_scrollable:
                w.xview(*args)

        h_scrollbar.config(command=sync_horizontal_scroll)
        v_scrollbar.config(command=sync_vertical_scroll)
        for f in horizontal_scrollable:
            f.canvas.config(xscrollcommand=h_scrollbar.set)
        for f in vertical_scrollable:
            f.canvas.config(yscrollcommand=v_scrollbar.set)

        # pack the outer frame
        outer_frame.pack(expand=1, fill="both")

    def scrollable_frame(self, outer_frame, row,col, width=None, height=None) -> Pane:
        sf = Pane(outer_frame)
        sf.grid(row=row, column=col, sticky='nsew')
        return sf


    def populate(self, columns, column_widths=150, num_rows=50 ):

        # header
        for i,c in enumerate(columns):
            tk.Label(self.header_frame.frame, text=c).grid(
                column=i,row=0,sticky='nsew', padx=1, pady=1)
            self.header_frame.frame.columnconfigure(i,minsize=column_widths, weight=0)

        # data frame has all entry boxes
        for r in range(num_rows):
            for c in range(len(columns)):
                tk.Entry(self.data_frame.frame).grid(column=c,row=r,sticky='nsew', padx=1, pady=1)
            self.data_frame.frame.rowconfigure(r, minsize=20)


        # titles
        for r in range(num_rows):
            tk.Label(self.title_frame.frame, text=f"Row {r}").grid(column=0,row=r,sticky='nsew', padx=1, pady=1)
            self.title_frame.frame.rowconfigure(r, minsize=20)

        # footer
        for c in range(len(columns)):
            tk.Label(self.footer_frame.frame, text=f"{c}").grid(column=c,row=2,sticky='nsew', padx=1, pady=1)
            self.footer_frame.frame.columnconfigure(c,minsize=column_widths, weight=0)

        # summary
        for r in range(num_rows):
            tk.Label(self.summary_frame.frame, text=f"Summary {r}").grid(column=0,row=r,sticky='nsew', padx=1, pady=1)
            self.summary_frame.frame.rowconfigure(r, minsize=20)






def main():
    root = tk.Tk()
    root.title("Employee List")
    columns = ("First Name","Last Name","Gender","City","Salary","Bonus")

    al = AllocationLayout(root)
    al.populate(columns,150)


    root.mainloop()


if __name__ == '__main__':
    main()


import pytest

from src.scheduling_and_allocation.model import Schedule, Course, Stream, Lab, Teacher, WeekDay


# ============================================================================
# tests
# ============================================================================
def test_constructor():
    schedule = Schedule()
    assert isinstance(schedule, Schedule)


def test_interface():
    schedule = Schedule()
    assert hasattr(schedule, "read_file")
    assert hasattr(schedule, "write_file")

    assert hasattr(schedule, 'add_update_course')
    assert hasattr(schedule, 'add_update_stream')
    assert hasattr(schedule, 'add_update_lab')
    assert hasattr(schedule, 'add_update_teacher')

    assert hasattr(schedule, 'labs')
    assert hasattr(schedule, 'courses')
    assert hasattr(schedule, 'teachers')
    assert hasattr(schedule, 'streams')

    assert hasattr(schedule, 'get_course_by_number')
    assert hasattr(schedule, 'get_lab_by_number')
    assert hasattr(schedule, 'get_stream_by_number')
    assert hasattr(schedule, 'get_teacher_by_name')

    assert hasattr(schedule, 'get_teacher_by_number')

    assert hasattr(schedule, 'remove_teacher')
    assert hasattr(schedule, 'remove_stream')
    assert hasattr(schedule, 'remove_lab')
    assert hasattr(schedule, 'remove_course')

    assert hasattr(schedule, "get_teachers_assigned_to_any_course")
    assert hasattr(schedule, 'get_streams_assigned_to_any_course')
    assert hasattr(schedule, 'get_labs_assigned_to_any_course')
    assert hasattr(schedule, 'get_courses_for_teacher')

    assert hasattr(schedule, 'sections')
    assert hasattr(schedule, 'blocks')

    assert hasattr(schedule, 'get_sections_for_teacher')
    assert hasattr(schedule, 'get_sections_for_stream')
    assert hasattr(schedule, 'get_blocks_for_teacher')
    assert hasattr(schedule, 'get_blocks_in_lab')
    assert hasattr(schedule, 'get_blocks_for_stream')
    assert hasattr(schedule, 'get_blocks_for_obj')

    assert hasattr(schedule, 'clear_all_from_course')
    assert hasattr(schedule, 'calculate_conflicts')


def test_add_course():
    s = Schedule()
    assert len(s._courses) == 0
    c1 = s.add_update_course('420-DBF')
    assert len(s._courses) == 1
    c2 = s.add_update_course('420-ABC')
    assert len(s._courses) == 2
    assert c1.number in s._courses
    assert c2.number in s._courses
    assert isinstance(c1, Course)


def test_add_stream():
    s = Schedule()
    assert len(s._streams) == 0
    c1 = s.add_update_stream('4A')
    assert len(s._streams) == 1
    c2 = s.add_update_stream('4B')
    assert len(s._streams) == 2
    assert c1.number in s._streams
    assert c2.number in s._streams
    assert isinstance(c1, Stream)


def test_add_lab():
    s = Schedule()
    assert len(s._labs) == 0
    c1 = s.add_update_lab('4A')
    assert len(s._labs) == 1
    c2 = s.add_update_lab('4B')
    assert len(s._labs) == 2
    assert c1.number in s._labs
    assert c2.number in s._labs
    assert isinstance(c1, Lab)


def test_add_teacher():
    s = Schedule()
    assert len(s._teachers) == 0
    c1 = s.add_update_teacher('4A', 'Doe')
    assert len(s._teachers) == 1
    c2 = s.add_update_teacher('4B', 'Doe')
    assert len(s._teachers) == 2
    assert c1.number in s._teachers
    assert c2.number in s._teachers
    assert isinstance(c1, Teacher)


def test_return_courses_as_tuple():
    s = Schedule()
    assert len(s.courses()) == 0
    o1 = Course('ABC')
    o2 = Course('ABD')
    s._courses = {o1.number: o1, o2.number: o2}
    assert len(s.courses()) == 2
    assert isinstance(s.courses(), tuple)
    assert o1 in s.courses()
    assert o2 in s.courses()


def test_return_labs_as_tuple():
    s = Schedule()
    assert len(s.labs()) == 0
    o1 = Lab('ABC')
    o2 = Lab('ABD')
    s._labs = {o1.number: o1, o2.number: o2}
    assert len(s.labs()) == 2
    assert isinstance(s.labs(), tuple)
    assert o1 in s.labs()
    assert o2 in s.labs()


def test_return_streams_as_tuple():
    s = Schedule()
    assert len(s.streams()) == 0
    o1 = Stream('ABC')
    o2 = Stream('ABD')
    s._streams = {o1.number: o1, o2.number: o2}
    assert len(s.streams()) == 2
    assert isinstance(s.streams(), tuple)
    assert o1 in s.streams()
    assert o2 in s.streams()


def test_return_teachers_as_tuple():
    s = Schedule()
    assert len(s.teachers()) == 0
    o1 = Teacher('ABC', 'Doe')
    o2 = Teacher('DEF', 'Doe')
    s._teachers = {o1.number: o1, o2.number: o2}
    assert len(s.teachers()) == 2
    assert isinstance(s.teachers(), tuple)
    assert o1 in s.teachers()
    assert o2 in s.teachers()


def test_get_course_by_number():
    s = Schedule()
    o1 = Course('ABC')
    o2 = Course('DEF')
    s._courses = {o1.number: o1, o2.number: o2}
    assert o2 == s.get_course_by_number('DEF')
    assert o1 == s.get_course_by_number('ABC')
    assert s.get_course_by_number("boo") is None


def test_get_lab_by_number():
    s = Schedule()
    o1 = Lab('ABC')
    o2 = Lab('DEF')
    s._labs = {o1.number: o1, o2.number: o2}
    assert o2 == s.get_lab_by_number('DEF')
    assert o1 == s.get_lab_by_number('ABC')
    assert s.get_lab_by_number("boo") is None


def test_get_stream_by_number():
    s = Schedule()
    o1 = Stream('ABC')
    o2 = Stream('DEF')
    s._streams = {o1.number: o1, o2.number: o2}
    assert o2 == s.get_stream_by_number('DEF')
    assert o1 == s.get_stream_by_number('ABC')
    assert s.get_stream_by_number("boo") is None


def test_get_teacher_by_name():
    s = Schedule()
    o1 = Teacher('Jane', 'Doe')
    o2 = Teacher('John', 'Doe')
    s._teachers = {o1.number: o1, o2.number: o2}
    assert o2 == s.get_teacher_by_name('John', 'Doe')
    assert o2 == s.get_teacher_by_name('JoHn', 'dOe')
    assert o1 == s.get_teacher_by_name('Jane', 'Doe')
    assert s.get_teacher_by_name("Jane", "boo") is None


def test_get_teacher_by_number():
    s = Schedule()
    o1 = Teacher('Jane', 'Doe')
    o2 = Teacher('John', 'Doe')
    s._teachers = {o1.number: o1, o2.number: o2}
    assert o2 == s.get_teacher_by_number(o2.number)
    assert o1 == s.get_teacher_by_number(o1.number)
    assert s.get_teacher_by_number("666666") is None


def test_get_teacher_by_id():
    s = Schedule()
    o1 = Teacher('Jane', 'Doe')
    o2 = Teacher('John', 'Doe')
    s._teachers = {o1.number: o1, o2.number: o2}
    assert o2 == s.get_teacher_by_number(o2.number)
    assert o1 == s.get_teacher_by_number(o1.number)


def test_remove_course():
    s = Schedule()
    o1 = Course('ABC')
    o2 = Course('DEF')
    s._courses = {o1.number: o1, o2.number: o2}
    s.remove_course(o2)
    assert len(s._courses) == 1
    assert o2 not in s._courses.values()

    # try to remove non-existent course
    s.remove_course(Course("XYZ"))
    assert True


def test_remove_teacher_simple():
    s = Schedule()
    o1 = Teacher('ABC', 'Doe')
    o2 = Teacher('DEF', 'Doe')
    s._teachers = {o1.number: o1, o2.number: o2}
    s.remove_teacher(o2)
    assert len(s._teachers) == 1
    assert o2 not in s._teachers.values()

    # try to remove non-existent teacher
    s.remove_teacher(Teacher("XYZ", 'Doe'))
    assert True


def test_remove_teacher_from_courses():
    s = Schedule()
    o1 = Teacher('ABC', 'Doe')
    o2 = Teacher('DEF', 'Doe')
    s._teachers = {o1.number: o1, o2.number: o2}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("C2")
    # NOTE: adding teachers is meaningless unless there are sections in the course
    # NOTE: adding teachers is meaningless unless there are blocks in the section
    st1 = c1.add_section("ABC")
    st2 = c2.add_section("DEF")
    st1.add_block()
    st2.add_block()
    c1.add_teacher(o1)
    c1.add_teacher(o2)
    c2.add_teacher(o1)
    c2.add_teacher(o2)

    s.remove_teacher(o2)
    assert len(s._teachers) == 1
    assert o2 not in c1.teachers()
    assert o2 not in c2.teachers()
    assert o1 in c1.teachers()
    assert o1 in c2.teachers()


def test_remove_lab_simple():
    s = Schedule()
    o1 = Lab('ABC')
    o2 = Lab('DEF')
    s._labs = {o1.number: o1, o2.number: o2}
    s.remove_lab(o2)
    assert len(s._labs) == 1
    assert o2 not in s._teachers.values()

    # try to remove non-existent lab
    s.remove_lab(Lab("XYZ"))
    assert True


def test_remove_lab_complex():
    s = Schedule()
    o1 = Lab('ABC')
    o2 = Lab('DEF')
    s._labs = {o1.number: o1, o2.number: o2}

    c1 = s.add_update_course("C1")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    b1 = s1.add_block(WeekDay.get_from_string('MOnday'), 9.0, 1)
    b2 = s1.add_block(WeekDay.get_from_string('MONDAY'), 10.0, 1)
    b3 = s2.add_block(WeekDay.get_from_string('mon'), 10.0, 1)
    s1.add_lab(o2)
    b3.add_lab(o2)
    c1.add_lab(o1)

    assert o2 in b1.labs()
    assert o2 in b2.labs()
    assert o2 in b3.labs()
    assert o2 in s1.labs()
    assert o2 in s2.labs()
    s.remove_lab(o2)
    assert len(s._labs) == 1
    assert o2 not in b1.labs()
    assert o2 not in b2.labs()
    assert o2 not in b3.labs()
    assert o2 not in s1.labs()
    assert o2 not in s2.labs()


def test_remove_stream_simple():
    s = Schedule()
    o1 = Stream('ABC')
    o2 = Stream('DEF')
    s._streams = {o1.number: o1, o2.number: o2}
    s.remove_stream(o2)
    assert len(s._streams) == 1
    assert o2 not in s._streams.values()

    # try to remove non-existent stream
    s.remove_stream(Stream("XYZ"))
    assert True


def test_remove_stream_from_courses():
    s = Schedule()
    o1 = Stream('ABC')
    o2 = Stream('DEF')
    s._streams = {o1.number: o1, o2.number: o2}

    c1 = s.add_update_course("C1")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s1.add_stream(o1)
    s2.add_stream(o2)

    assert o2 not in s1.streams()
    assert o2 in s2.streams()
    assert o2 in c1.streams()
    s.remove_stream(o2)
    assert len(s._streams) == 1
    assert o2 not in s1.streams()
    assert o2 not in s2.streams()
    assert o2 not in c1.streams()


def test_get_teachers_assigned_to_any_course():
    s = Schedule()
    o1 = Teacher('ABC', 'Doe')
    o2 = Teacher('DEF', 'Doe')
    o3 = Teacher('XYZ', 'Doe')
    s._teachers = {o1.number: o1, o2.number: o2, o3.number: o3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("C2")
    # NOTE: adding teachers is meaningless unless there are sections in the course
    st1 = c1.add_section("ABC")
    st2 = c2.add_section("DEF")
    st1.add_block()
    st2.add_block()
    c1.add_teacher(o1)
    c1.add_teacher(o2)
    c2.add_teacher(o1)

    assert len(s.teachers()) == 3
    assert o3 in s.teachers()
    assert o3 not in s.get_teachers_assigned_to_any_course()
    assert o2 in s.get_teachers_assigned_to_any_course()
    assert o1 in s.get_teachers_assigned_to_any_course()
    assert len(s.get_teachers_assigned_to_any_course()) == 2


def get_streams_assigned_to_any_course():
    s = Schedule()
    o1 = Stream('ABC')
    o2 = Stream('DEF')
    o3 = Stream('XYZ')
    s._streams = {o1.number: o1, o2.number: o2, o3.number: o3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("C2")
    c1.add_stream(o1)
    c1.add_stream(o2)
    c2.add_stream(o1)

    assert len(s.streams()) == 3
    assert o3 in s.streams()
    assert o3 not in s.get_streams_assigned_to_any_course()
    assert o2 in s.get_streams_assigned_to_any_course()
    assert o1 in s.get_streams_assigned_to_any_course()
    assert len(s.get_streams_assigned_to_any_course()) == 2


def get_labs_assigned_to_any_course():
    s = Schedule()
    o1 = Lab('ABC')
    o2 = Lab('DEF')
    o3 = Lab('XYZ')
    s._labs = {o1.number: o1, o2.number: o2, o3.number: o3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("C2")
    c1.add_lab(o1)
    c1.add_lab(o2)
    c2.add_lab(o1)

    assert len(s.labs()) == 3
    assert o3 in s.labs
    assert o3 not in s.get_labs_assigned_to_any_course()
    assert o2 in s.get_labs_assigned_to_any_course()
    assert o1 in s.get_labs_assigned_to_any_course()
    assert len(s.get_labs_assigned_to_any_course()) == 2


def get_courses_for_teacher():
    s = Schedule()
    o1 = Teacher('ABC', 'Doe')
    o2 = Teacher('DEF', 'Doe')
    o3 = Teacher('XYZ', 'Doe')
    s._teachers = {o1.number: o1, o2.number: o2, o3.number: o3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("C2")
    c3 = s.add_update_course("C3")
    c1.add_teacher(o1)
    c1.add_teacher(o2)
    c2.add_teacher(o1)

    assert c1 in s.get_courses_for_teacher(o1)
    assert c2 in s.get_courses_for_teacher(o1)
    assert c3 not in s.get_courses_for_teacher(o1)
    assert len(s.get_courses_for_teacher(o1)) == 2
    assert c1 in s.get_courses_for_teacher(o2)
    assert c2 not in s.get_courses_for_teacher(o2)
    assert len(s.get_courses_for_teacher(o2)) == 1
    assert c3 not in s.get_courses_for_teacher(o2)


def test_sections():
    s = Schedule()
    o1 = Lab('ABC')
    o2 = Lab('DEF')
    s._labs = {o1.number: o1, o2.number: o2}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("C2")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s3 = c2.add_section("1")
    s4 = c2.add_section("2")

    assert len(s.sections()) == 4
    assert s1 in s.sections()
    assert s2 in s.sections()
    assert s3 in s.sections()
    assert s4 in s.sections()


def test_blocks():
    s = Schedule()
    o1 = Lab('ABC')
    o2 = Lab('DEF')
    s._labs = {o1.number: o1, o2.number: o2}

    c1 = s.add_update_course("C1")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    sections = s.sections()
    b1 = s1.add_block(WeekDay.Monday, 9.0, 1)
    b2 = s1.add_block(WeekDay.Monday, 10.0, 1)
    b3 = s2.add_block(WeekDay.Monday, 10.0, 1)

    assert len(s.blocks()) == 3
    x = s.blocks()
    assert b1 in s.blocks()
    assert b2 in s.blocks()
    assert b3 in s.blocks()


def test_get_sections_for_teacher():
    s = Schedule()
    l1 = Lab('ABC')
    l2 = Lab('DEF')
    s._labs = {l1.number: l1, l2.number: l2}
    st1 = Stream('ABC')
    st2 = Stream('DEF')
    s._streams = {st1.number: st1, st2.number: st2}
    t1 = Teacher("ABC", "Doe")
    t2 = Teacher("DEF", "Doe")
    t3 = Teacher("XYZ", "Doe")
    s._teachers = {t1.number: t1, t2.number: t2, t3.number: t3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("c3")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s3 = c2.add_section("1")
    s4 = c2.add_section("2")

    b1 = s1.add_block(WeekDay.Monday, 9.0, 1)
    b2 = s1.add_block(WeekDay.Monday, 10.0, 1)
    b3 = s2.add_block(WeekDay.Monday, 10.0, 1)
    b4 = s3.add_block(WeekDay.Monday, 9.0, 1)
    b5 = s3.add_block(WeekDay.Monday, 10.0, 1)
    b6 = s4.add_block(WeekDay.Monday, 10.0, 1)

    c1.add_teacher(t1)
    c2.add_teacher(t2)

    s1.add_stream(st1)
    s3.add_stream(st1)
    s2.add_stream(st2)
    s4.add_stream(st2)

    c1.add_lab(l1)
    c2.add_lab(l2)

    assert len(s.get_sections_for_teacher(t1)) == 2
    assert s1 in s.get_sections_for_teacher(t1)
    assert s2 in s.get_sections_for_teacher(t1)
    assert isinstance(s.get_sections_for_teacher(t1), tuple)
    assert len(s.get_sections_for_teacher(t3)) == 0


def test_get_sections_for_stream():
    s = Schedule()
    l1 = Lab('ABC')
    l2 = Lab('DEF')
    s._labs = {l1.number: l1, l2.number: l2}
    st1 = Stream('ABC')
    st2 = Stream('DEF')
    s._streams = {st1.number: st1, st2.number: st2}
    t1 = Teacher("ABC", "Doe")
    t2 = Teacher("DEF", "Doe")
    t3 = Teacher("XYZ", "Doe")
    s._teachers = {t1.number: t1, t2.number: t2, t3.number: t3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("c3")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s3 = c2.add_section("1")
    s4 = c2.add_section("2")

    b1 = s1.add_block(WeekDay.Monday, 9.0, 1)
    b2 = s1.add_block(WeekDay.Monday, 10.0, 1)
    b3 = s2.add_block(WeekDay.Monday, 10.0, 1)
    b4 = s3.add_block(WeekDay.Monday, 9.0, 1)
    b5 = s3.add_block(WeekDay.Monday, 10.0, 1)
    b6 = s4.add_block(WeekDay.Monday, 10.0, 1)

    c1.add_teacher(t1)
    c2.add_teacher(t2)

    s1.add_stream(st1)
    s3.add_stream(st1)
    s2.add_stream(st2)
    s4.add_stream(st2)

    c1.add_lab(l1)
    c2.add_lab(l2)

    assert len(s.get_sections_for_stream(st2)) == 2
    assert s2 in s.get_sections_for_stream(st2)
    assert s4 in s.get_sections_for_stream(st2)
    assert s1 not in s.get_sections_for_stream(st2)


def test_get_blocks_for_teacher():
    s = Schedule()
    l1 = Lab('ABC')
    l2 = Lab('DEF')
    s._labs = {l1.number: l1, l2.number: l2}
    st1 = Stream('ABC')
    st2 = Stream('DEF')
    s._streams = {st1.number: st1, st2.number: st2}
    t1 = Teacher("ABC", "Doe")
    t2 = Teacher("DEF", "Doe")
    t3 = Teacher("XYZ", "Doe")
    s._teachers = {t1.number: t1, t2.number: t2, t3.number: t3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("c3")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s3 = c2.add_section("1")
    s4 = c2.add_section("2")

    b1 = s1.add_block(WeekDay.Monday, 9.0, 1)
    b2 = s1.add_block(WeekDay.Monday, 10.0, 1)
    b3 = s2.add_block(WeekDay.Monday, 10.0, 1)
    b4 = s3.add_block(WeekDay.Monday, 9.0, 1)
    b5 = s3.add_block(WeekDay.Monday, 10.0, 1)
    b6 = s4.add_block(WeekDay.Monday, 10.0, 1)

    c1.add_teacher(t1)
    c2.add_teacher(t2)

    s1.add_stream(st1)
    s3.add_stream(st1)
    s2.add_stream(st2)
    s4.add_stream(st2)

    c1.add_lab(l1)
    c2.add_lab(l2)

    assert len(s.get_blocks_for_teacher(t1)) == 3
    assert b1 in s.get_blocks_for_teacher(t1)
    assert b2 in s.get_blocks_for_teacher(t1)
    assert b3 in s.get_blocks_for_teacher(t1)
    assert len(s.get_blocks_for_teacher(t3)) == 0


def test_get_blocks_in_lab():
    s = Schedule()
    l1 = Lab('ABC')
    l2 = Lab('DEF')
    l3 = Lab('XYZ')
    s._labs = {l1.number: l1, l2.number: l2}
    st1 = Stream('ABC')
    st2 = Stream('DEF')
    s._streams = {st1.number: st1, st2.number: st2}
    t1 = Teacher("ABC", "Doe")
    t2 = Teacher("DEF", "Doe")
    t3 = Teacher("XYZ", "Doe")
    s._teachers = {t1.number: t1, t2.number: t2, t3.number: t3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("c3")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s3 = c2.add_section("1")
    s4 = c2.add_section("2")

    b1 = s1.add_block(WeekDay.Monday, 9.0, 1)
    b2 = s1.add_block(WeekDay.Monday, 10.0, 1)
    b3 = s2.add_block(WeekDay.Monday, 10.0, 1)
    b4 = s3.add_block(WeekDay.Monday, 9.0, 1)
    b5 = s3.add_block(WeekDay.Monday, 10.0, 1)
    b6 = s4.add_block(WeekDay.Monday, 10.0, 1)

    c1.add_teacher(t1)
    c2.add_teacher(t2)

    s1.add_stream(st1)
    s3.add_stream(st1)
    s2.add_stream(st2)
    s4.add_stream(st2)

    c1.add_lab(l1)
    c2.add_lab(l2)

    assert len(s.get_blocks_in_lab(l1)) == 3
    assert b1 in s.get_blocks_in_lab(l1)
    assert b2 in s.get_blocks_in_lab(l1)
    assert b3 in s.get_blocks_in_lab(l1)
    assert isinstance(s.get_blocks_in_lab(l1), tuple)
    assert len(s.get_blocks_in_lab(l3)) == 0


def test_get_blocks_for_stream():
    s = Schedule()
    l1 = Lab('ABC')
    l2 = Lab('DEF')
    l3 = Lab('XYZ')
    s._labs = {l1.number: l1, l2.number: l2}
    st1 = Stream('ABC')
    st2 = Stream('DEF')
    st3 = Stream("XYZ")
    s._streams = {st1.number: st1, st2.number: st2}
    t1 = Teacher("ABC", "Doe")
    t2 = Teacher("DEF", "Doe")
    t3 = Teacher("XYZ", "Doe")
    s._teachers = {t1.number: t1, t2.number: t2, t3.number: t3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("c3")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s3 = c2.add_section("1")
    s4 = c2.add_section("2")

    b1 = s1.add_block(WeekDay.Monday, 9.0, 1)
    b2 = s1.add_block(WeekDay.Monday, 10.0, 1)
    b3 = s2.add_block(WeekDay.Monday, 11.0, 1)
    b4 = s3.add_block(WeekDay.Tuesday, 9.0, 1)
    b5 = s3.add_block(WeekDay.Tuesday, 10.0, 1)
    b6 = s4.add_block(WeekDay.Tuesday, 11.0, 1)

    c1.add_teacher(t1)
    c2.add_teacher(t2)

    s1.add_stream(st1)
    s3.add_stream(st1)
    s2.add_stream(st2)
    s4.add_stream(st2)

    c1.add_lab(l1)
    c2.add_lab(l2)

    assert len(s.get_blocks_for_stream(st1)) == 4
    assert b1 in s.get_blocks_for_stream(st1)
    assert b2 in s.get_blocks_for_stream(st1)
    assert b4 in s.get_blocks_for_stream(st1)
    assert b5 in s.get_blocks_for_stream(st1)

    assert isinstance(s.get_blocks_for_stream(st1), tuple)
    assert len(s.get_blocks_for_stream(st3)) == 0


def test_get_blocks_for_obj():
    s = Schedule()
    l1 = Lab('ABC')
    l2 = Lab('DEF')
    l3 = Lab('XYZ')
    s._labs = {l1.number: l1, l2.number: l2}
    st1 = Stream('ABC')
    st2 = Stream('DEF')
    st3 = Stream("XYZ")
    s._streams = {st1.number: st1, st2.number: st2, st3.number: st3}
    t1 = Teacher("ABC", "Doe")
    t2 = Teacher("DEF", "Doe")
    t3 = Teacher("XYZ", "Doe")
    s._teachers = {t1.number: t1, t2.number: t2, t3.number: t3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("c3")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s3 = c2.add_section("1")
    s4 = c2.add_section("2")

    b1 = s1.add_block(WeekDay.Monday, 9.0, 1)
    b2 = s1.add_block(WeekDay.Monday, 10.0, 1)
    b3 = s2.add_block(WeekDay.Monday, 10.0, 1)
    b4 = s3.add_block(WeekDay.Monday, 9.0, 1)
    b5 = s3.add_block(WeekDay.Monday, 10.0, 1)
    b6 = s4.add_block(WeekDay.Monday, 10.0, 1)

    c1.add_teacher(t1)
    c2.add_teacher(t2)

    s1.add_stream(st1)
    s3.add_stream(st1)
    s2.add_stream(st2)
    s4.add_stream(st2)

    c1.add_lab(l1)
    c2.add_lab(l2)

    assert s.get_blocks_for_teacher(t1) == s.get_blocks_for_obj(t1)
    assert s.get_blocks_in_lab(l1) == s.get_blocks_for_obj(l1)
    assert s.get_blocks_for_stream(st1) == s.get_blocks_for_obj(st1)


def test_clear_all():
    s = Schedule()
    l1 = Lab('ABC')
    l2 = Lab('DEF')
    l3 = Lab('XYZ')
    s._labs = {l1.number: l1, l2.number: l2}
    st1 = Stream('ABC')
    st2 = Stream('DEF')
    st3 = Stream("XYZ")
    s._streams = {st1.number: st1, st2.number: st2, st3.number: st3}
    t1 = Teacher("ABC", "Doe")
    t2 = Teacher("DEF", "Doe")
    t3 = Teacher("XYZ", "Doe")
    s._teachers = {t1.number: t1, t2.number: t2, t3.number: t3}

    c1 = s.add_update_course("C1")
    c2 = s.add_update_course("c3")
    s1 = c1.add_section("1")
    s2 = c1.add_section("2")
    s3 = c2.add_section("1")
    s4 = c2.add_section("2")

    b1 = s1.add_block(WeekDay.Monday, 9.0, 1)
    b2 = s1.add_block(WeekDay.Monday, 10.0, 1)
    b3 = s2.add_block(WeekDay.Monday, 10.0, 1)
    b4 = s3.add_block(WeekDay.Monday, 9.0, 1)
    b5 = s3.add_block(WeekDay.Monday, 10.0, 1)
    b6 = s4.add_block(WeekDay.Monday, 10.0, 1)

    c1.add_teacher(t1)
    c2.add_teacher(t2)

    s1.add_stream(st1)
    s3.add_stream(st1)
    s2.add_stream(st2)
    s4.add_stream(st2)

    c1.add_lab(l1)
    c2.add_lab(l2)

    s.clear_all_from_course(c1)

    assert len(c1.teachers()) == 0
    assert len(c1.labs()) == 0
    assert len(c1.streams()) == 0


def test_calculate_conflicts():
    """too much trouble to test this right now"""
    # TODO: write this test
    assert True



from functools import partial
from tkinter import *

from src.scheduling_and_allocation.modified_tk import Scrolled


class AddRemove:
    def __init__(self, frame: Frame,
                 get_add_list,
                 get_remove_list,
                 to_add_function,
                 to_remove_function,
                 add_text="Add to",
                 remove_text="Remove from"):
        self.frame = frame
        self.to_add_function = to_add_function
        self.to_remove_function = to_remove_function
        self.add_text = add_text
        self.remove_text = remove_text
        self.add_listbox = None
        self.remove_listbox = None
        self.adds = []
        self.removes = []
        self.get_add_list = get_add_list
        self.get_remove_list = get_remove_list

        f = Frame(self.frame)
        f.grid(column=0, stick='nsew', row=0)
        Label(f, text=self.add_text).pack()
        sf = Frame(f)
        sf.pack(fill='both', expand=1)
        s: Scrolled = Scrolled(sf, 'Listbox', scrollbars='oe')
        self.add_listbox = s.widget
        self.add_listbox.configure(borderwidth="5", relief="sunken")
        s.widget.bind('<Button-1>', partial(self._cmd_double, 'add'))

        f = Frame(self.frame)
        f.grid(column=1, stick='nsew', row=0)
        Label(f, text=self.remove_text).pack()
        sf = Frame(f)
        sf.pack(fill='both', expand=1)
        s: Scrolled = Scrolled(sf, 'Listbox', scrollbars='oe')
        self.remove_listbox = s.widget
        self.remove_listbox.configure(borderwidth="5", relief="sunken")
        s.widget.bind('<Button-1>',  partial(self._cmd_double, 'remove'))

        self.refresh()

    def _cmd_double(self, which: str, e: Event):
        if which == "add":
            widget = self.add_listbox
            index = widget.nearest(e.y)
            if index < len(self.adds):
                self.to_add_function(self.adds[index])
        else:
            widget = self.add_listbox
            index = widget.nearest(e.y)
            if index < len(self.removes):
                self.to_remove_function(self.removes[index])
        self.refresh()

    def refresh(self):
        self.adds = self.get_add_list()
        self.removes = self.get_remove_list()
        self.add_listbox.delete(0, "end")
        self.remove_listbox.delete(0, "end")
        for item in self.adds:
            self.add_listbox.insert("end", str(item))
        for item in self.removes:
            self.remove_listbox.insert("end", str(item))


to_add = [1,3,5,7,11,13,15,17,19,21,23,25,27,29,31,33]
to_remove = [2,4,6,8]
def get_add_list():
    return to_add
def get_remove_list():
    return to_remove
def add_function(obj):
    to_remove.append(obj)
    to_remove.sort()
    to_add.remove(obj)
def remove_function(obj):
    to_add.append(obj)
    to_add.sort()
    to_remove.remove(obj)





mw=Toplevel()
frame = Frame(mw)
frame.pack()
mw.geometry("400x400")
AddRemove(frame, get_add_list,get_remove_list, add_function, remove_function)

mw.mainloop()




from dataclasses import dataclass
from tkinter import *
from typing import Optional

from src.scheduling_and_allocation.Utilities.id_generator import IdGenerator
from src.scheduling_and_allocation.gui_generics.menu_and_toolbars import MenuItem, MenuType
from src.scheduling_and_allocation.gui_pages.view_dynamic_tk import ViewDynamicTk
from src.scheduling_and_allocation.model import Lab, Stream, Teacher, ResourceType, Schedule, Block

mw = Tk()
frame = Frame(mw)
frame.pack(expand=1,fill='both')


RESOURCE = Lab | Stream | Teacher

_gui_block_ids = IdGenerator()

frame = Frame(mw)
frame.pack(expand=1, fill = 'both')
resource_type = ResourceType.teacher
schedule = Schedule("../unit_tests_presenter/silly_winter.csv")
gui_blocks: dict[str, Block] = {}
blocks: list[Block] = []

def draw_blocks():
    global blocks
    blocks = schedule.blocks()
    gui_blocks.clear()

    for block in blocks:
        gui_block_id: int = _gui_block_ids.get_new_id()
        gui_tag = f"gui_tag_{gui_block_id}"
        gui_blocks[gui_tag] = block

        view.draw_block(resource_type=resource_type,
                            day=block.day.value, start_time=block.start, duration=block.duration,
                            text=gui_tag,
                            gui_block_id=gui_tag,
                            movable=block.movable)
        view.colour_block(gui_tag, resource_type, block.movable, conflict=block.conflict)

def gui_block_is_moving(gui_id: str,  gui_block_day: float, gui_block_start_time: float):

    block: Block = gui_blocks.get(gui_id,None)
    block.start = gui_block_start_time
    block.snap_to_day(gui_block_day)
    block.snap_to_time()
    schedule.calculate_conflicts()
    for gui_tag, block in gui_blocks.items():
        view.colour_block(gui_tag, resource_type, is_movable=block.movable, conflict=block.conflict)

def gui_block_has_dropped(gui_id: str):
    print("block dropped")
    block: Block = gui_blocks.get(gui_id,None)
    view.move_gui_block(gui_id, block.day.value,block.start)


def tree_popup_menu( gui_id) -> Optional[list[MenuItem]]:
    menu_list: list[MenuItem] = [MenuItem(menu_type=MenuType.Command, label="Click me",
                                          command=lambda : print(f"menu for {gui_id}"))]
    return menu_list

# create the gui representation of the view
view = ViewDynamicTk( frame, "Testing View", ResourceType.teacher)

# Need to set handlers here, instead on constructor, so that it is easier for testing
view.get_popup_menu_handler = tree_popup_menu
view.refresh_blocks_handler = draw_blocks
view.on_closing_handler = lambda *_: print ("on closing")
view.double_click_block_handler = lambda x: print("double clicked",x)
view.gui_block_is_moving_handler = gui_block_is_moving
view.gui_block_has_dropped_handler = gui_block_has_dropped
view.undo_handler = lambda: print("undo")
view.redo_handler = lambda: print("redo")
view.draw()

mw.mainloop()


from __future__ import annotations

import os
import sys
import tkinter as tk
from tkinter.simpledialog import Dialog

bin_dir: str = os.path.dirname(os.path.realpath(__file__))
sys.path.append(os.path.join(bin_dir, "../../"))
import src.scheduling_and_allocation.modified_tk.InitGuiFontsAndColours as fac

mw = tk.Tk()
mw.geometry("400x400")

class ChangeFont(Dialog):
    def __init__(self, frame:tk.Frame, title="Change Font", font_size: int=10):
        self.warning_label = None
        self.label_big = None
        self.label_normal = None
        self.label_small = None
        self.font_size_tk = tk.StringVar(value=str(font_size))
        self.fonts = fac.TkFonts(frame.winfo_toplevel(), font_size)

        super().__init__(frame.winfo_toplevel(), title)

    # ================================================================================================================
    # The content of the main body of the dialog box
    # ================================================================================================================
    def body(self, frame:tk.Frame):
        """
        :param frame: the frame where you are gonna put stuff
        """
        subframe = tk.Frame(frame)
        self.frame=frame

        self.label_small:tk.Label = tk.Label(frame, text="Small text: ABC abc 0123", font=self.fonts.small)
        self.label_normal:tk.Label = tk.Label(frame, text="Normal text: ABC abc 0123", font=self.fonts.normal)
        self.label_big:tk.Label = tk.Label(frame, text="Big text: ABC abc 0123", font=self.fonts.big)
        self.warning_label = tk.Label(frame, text="Must restart application for font changes to take effect!!",
                 font=self.fonts.bigbold)

        modify_font = frame.winfo_toplevel().register(self._change_font)
        font_size_label=tk.Label(subframe, text="Font Size:")
        font_size_entry = tk.Entry(subframe, textvariable=self.font_size_tk,
                                   validate='key',
                                   validatecommand=(modify_font, '%P', '%s')
                                   )

        # layout
        self.warning_label.pack(side="top", expand=1, fill='x', pady=20, padx=10)
        subframe.pack(side="top", expand=1, fill='x', pady=20)
        font_size_label.pack(side='left', expand=1, fill='x', padx=10)
        font_size_entry.pack(side='left', expand=1, fill='x', padx=10)

        self.label_small.pack(side='top', expand=1, fill='x', padx=20)
        self.label_normal.pack(side='top', expand=1, fill='x', padx=20)
        self.label_big.pack(side='top', expand=1, fill='x', padx=20)

        return font_size_entry

    def _change_font(self, number: str, *_):
        if number == "":
            return True
        try:
            int(number)

            self.fonts = fac.TkFonts(frame.winfo_toplevel(), int(number))
            self.warning_label.config(font=self.fonts.bigbold)
            self.label_big.config(font=self.fonts.big)
            self.label_small.config(font=self.fonts.small)
            self.label_normal.config(font=self.fonts.normal)
            return True
        except ValueError:
            return False

    # ================================================================================================================
    # validate before applying
    # ================================================================================================================
    def validate(self):
        """
        Is the data, as entered by the user, valid?
        :return: True if data is good (changes are applied), false otherwise (nothing happens)
        """
        return True

    # ================================================================================================================
    # apply changes
    # ================================================================================================================
    def apply(self):
        """apply the changes and close the dialog"""
        pass


frame = tk.Frame(mw)
tk.Label(frame,text="Sample Text").pack()
frame.pack(expand=1, fill='both')
dg = ChangeFont(frame)

mw.mainloop()



"""
# ============================================================================
# Create a list of all resources
# - edit properties
# - create new resource
# - delete resource
#
# Events triggered by EditResourcesTk
#       delete_obj(row_data)
#       save(all_data)
#
# ============================================================================
"""

from typing import Optional, Callable

from ..Utilities import Preferences
from ..model import Schedule, ResourceType
from ..gui_pages import EditResourcesTk, DEColumnDescription

property_conversions_from_str = {
    "firstname": lambda x: x,
    "lastname": lambda x: x,
    "release": lambda x: float(x) if x else 0,
    "teacher_id": lambda x: x,
    "description": lambda x: x,
    "number": lambda x: x,
}


# =====================================================================================================================
# Class Edit Resources
# =====================================================================================================================
class EditResources:

    # -----------------------------------------------------------------------------------------------------------------
    # constructor
    # -----------------------------------------------------------------------------------------------------------------
    def __init__(self,
                 dirty_flag_method: Callable[[Optional[bool]], bool],
                 frame,
                 view_type: ResourceType,
                 schedule: Optional[Schedule],
                 preferences: Preferences = None,
                 gui=None):
        """
        Creates the basic EditResources (a simple matrix)
        :dirty_flag_method: function that gets/sets the dirty flag
        :param frame: gui container object
        :param view_type: Is it a lab, stream or teacher
        :param schedule: The Schedule object
        :gui: the gui object that shows data (optional)
        """

        self.preferences = preferences
        # Resource type specifics
        match view_type:
            case ResourceType.teacher:
                self.column_descriptions: list[DEColumnDescription] = [
                    DEColumnDescription(title="ID", width=5, property="teacher_id", unique_id=True),
                    DEColumnDescription(title="First Name", width=15, property="firstname", unique_id=False),
                    DEColumnDescription(title="Last Name", width=15, property="lastname", unique_id=False),
                    DEColumnDescription(title="RT", width=8, property="release", unique_id=False),
                ]
                self._get_resources = schedule.teachers
                self._get_resource_by_number = schedule.get_teacher_by_number
                self._delete_resource = schedule.remove_teacher
                self._update_resource = schedule.add_update_teacher

            case ResourceType.lab:
                self.column_descriptions: list[DEColumnDescription] = [
                    DEColumnDescription(title="Room", width=14, property="number", unique_id=True),
                    DEColumnDescription(title="Description", width=50, property="description", unique_id=False),
                ]
                self._get_resources = schedule.labs
                self._get_resource_by_number = schedule.get_lab_by_number
                self._delete_resource = schedule.remove_lab
                self._update_resource = schedule.add_update_lab

            case ResourceType.stream:
                self.column_descriptions: list[DEColumnDescription] = [
                    DEColumnDescription(title="Room", width=14, property="number", unique_id=True),
                    DEColumnDescription(title="Description", width=50, property="description", unique_id=False),
                ]
                self._get_resources = schedule.streams
                self._get_resource_by_number = schedule.get_stream_by_number
                self._delete_resource = schedule.remove_stream
                self._update_resource = schedule.add_update_stream

            case _:
                self.column_descriptions: list[DEColumnDescription] = [
                    DEColumnDescription(title="", width=10, property="", unique_id=True),
                ]
                self._get_resources = lambda: []
                self._get_resource_by_number = lambda x: None
                self._delete_resource = lambda x: None
                self._update_resource = lambda *args, **kwargs: None

        # other initializations
        if not gui:
            self.gui = EditResourcesTk(
                parent=frame,
                event_delete_handler=self.delete_obj,
                event_save_handler=self.save,
                preferences=self.preferences,
            )
        else:
            self.gui = gui

        disabled_columns = [True,] if view_type == ResourceType.teacher else []

        self.dirty_flag_method: Callable[[Optional[bool]], bool] = dirty_flag_method
        self.delete_queue: list = list()
        self.view_type: ResourceType = view_type
        self.schedule: Optional[Schedule] = schedule
        self.gui.initialize_columns(self.column_descriptions, disabled_columns)

    # -----------------------------------------------------------------------------------------------------------------
    # refresh the tables
    # -----------------------------------------------------------------------------------------------------------------
    def refresh(self):
        """update the table to represent the 'new' data"""

        # get the data
        objs = self._get_resources()

        # Create a 2-d array of the data that needs to be displayed.
        data = [[str(getattr(obj, col.property, None)) for col in self.column_descriptions] for obj in objs]

        # refresh the GUI_Pages
        self.gui.refresh(data)

        # purge the delete queue
        self.delete_queue.clear()

    # -----------------------------------------------------------------------------------------------------------------
    # Save updated data
    # -----------------------------------------------------------------------------------------------------------------
    def save(self, all_data: list[list[str]]):
        """
        Save any changes that the user entered in the GUI_Pages form.
        :param all_data: a 2-d array of string values in columns and rows
        """

        needs_updating = self._check_for_changed_resources(all_data)

        # do we need to change anything?
        changes = False
        changes = len(needs_updating) or changes
        changes = len(self.delete_queue) or changes
        if not changes:
            return

        # Update or add changes.
        for obj in self.delete_queue:
            self._delete_resource(obj)
        for init_arguments in needs_updating:
            self._update_resource(**init_arguments)

        # all done saving
        self.dirty_flag_method(changes)
        self.refresh()

    # -----------------------------------------------------------------------------------------------------------------
    # delete object
    # -----------------------------------------------------------------------------------------------------------------
    def delete_obj(self, data: list[str], *_):
        """Save delete requests, to be processed later. Note that all data is given in string representation"""
        obj = None
        for index, column in enumerate(self.column_descriptions):
            if column.unique_id and data[index]:
                unique_id = property_conversions_from_str[column.property](data[index])
                obj = self.schedule.get_view_type_obj_by_id(self.view_type, unique_id)
                break
        if obj:
            self.delete_queue.append(obj)

    # -----------------------------------------------------------------------------------------------------------------
    # Check if any of the resources have been modified
    # -----------------------------------------------------------------------------------------------------------------
    def _check_for_changed_resources(self, all_data) -> list[dict]:
        """Go through data that is in gui, and compare to existing objects
        :return dictionary: key/value pairs required for this resource type's constructor
        """

        changed_objects: list[dict] = []

        # Read data from the data object (ignore lines with no data in them).
        for data in (d for d in all_data if len([x for x in d if x])):

            obj = None
            constructor_inputs = dict()

            # get all inputs required for the constructor
            for index, column in enumerate(self.column_descriptions):
                if column.unique_id and data[index]:
                    unique_id = property_conversions_from_str[column.property](data[index])
                    obj = self.schedule.get_view_type_obj_by_id(self.view_type, unique_id)
                    constructor_inputs[column.property] = unique_id
                elif not column.unique_id:
                    if index < len(data):
                        constructor_inputs[column.property] = property_conversions_from_str[column.property](
                            data[index])

            # if object is modified or new, save constructor inputs
            if obj is not None and self._object_changed(obj, data):
                changed_objects.append(constructor_inputs)
            if obj is None:
                changed_objects.append(constructor_inputs)

        return changed_objects

    # -----------------------------------------------------------------------------------------------------------------
    # does an object need to be updated?
    # -----------------------------------------------------------------------------------------------------------------
    def _object_changed(self, obj, data) -> bool:
        """compares each column data with the attribute for the object and returns true if any of them are different"""

        for col, column in enumerate(self.column_descriptions):
            if hasattr(obj, column.property):
                if str(getattr(obj, column.property, None)) != data[col]:
                    return True
        return False



from tkinter import *

from src.scheduling_and_allocation.gui_generics.menu_and_toolbars import generate_menu, make_toolbar, ToolbarItem, \
    MenuItem, MenuType
from src.scheduling_and_allocation.modified_tk import set_default_fonts_and_colours


# =================================================================================================
# Test the GUI_Pages menu and _toolbar creation work as required
# =================================================================================================


def main():
    mw = Tk()
    colours, fonts = set_default_fonts_and_colours(mw, invert=True)

    # preparation
    (buttons, toolbar_info, menu_details) = define_inputs()

    menu_bar = Menu(mw)
    generate_menu(mw, menu_details, menu_bar)
    mw.configure(menu=menu_bar)

    # action
    toolbar = make_toolbar(mw, buttons, toolbar_info)
    toolbar.pack(side='top', expand=0, fill='x')

    toolbar2 = make_toolbar(mw, buttons, toolbar_info, colours, bg="blue")
    toolbar2.pack(side='top', expand=0, fill='x')

    # test (manually)
    Label(mw, anchor='w', text="").pack()
    Label(mw, anchor='w', text="").pack()
    Label(mw, anchor='w', text="Verify that there are TWO toolbars").pack(fill='x')
    Label(mw, anchor='w', text="The top _toolbar should have the same background Colour has the main window").pack(
        fill='x')
    Label(mw, anchor='w', text="The 2nd _toolbar should have a background Colour of blue").pack(fill='x')
    Label(mw, anchor='w', text="The 2nd _toolbar buttons should turn pink if mouse if hovered over the button").pack(
        fill='x')
    Label(mw, anchor='w', text="*** KNOWN BUG: hoverbg does not work in Tk/Toolbar.py").pack(fill='x')
    Label(mw, text="").pack()
    Label(mw, anchor='w', text='Verify that there are two file menu items, "file" and "print"').pack(fill='x')
    Label(mw, anchor='w', text='Verify that "file" has "New", "Open", "Save", "Save As", "Exit"').pack(fill='x')
    Label(mw, anchor='w', text='Verify that "print" has two sub-menus "pdf" and "latex"').pack(fill='x')
    Label(mw, anchor='w', text='Verify the "pdf" and "latex" have "Teacher/Lab/Stream" menu choices').pack(fill='x')
    Label(mw, anchor='w', text="").pack(fill='x')
    Label(mw, anchor='w', text="Toolbar should have three buttons, with a '|' between open and save").pack(fill='x')
    Label(mw, anchor='w', text="").pack(fill='x')
    Label(mw, anchor='w', text="Verify that all buttons and menu items print appropriate message to console").pack(
        fill='x')
    Label(mw, anchor='w', text="Verify that all menu item shortcut keys work as expected").pack(fill='x')
    Label(mw, anchor='w', text="If MAC, verify that all menu item 'command'-? shortcut keys work as expected").pack(
        fill='x')

    mw.mainloop()


# =================================================================================================
# Setup
# =================================================================================================
def define_inputs() -> tuple[list[str], dict[str, ToolbarItem], list[MenuItem]]:
    menu = list()

    # -----------------------------------------------------------------------------------------
    # File menu
    # -----------------------------------------------------------------------------------------
    file_menu = MenuItem(name='file', menu_type=MenuType.Cascade, label='File')
    file_menu.add_child(MenuItem(name='new', menu_type=MenuType.Command, label='New', accelerator='Ctrl-n',
                                 command=lambda *_: print(f"'File/New' selected {_}")))
    file_menu.add_child(MenuItem(name='open', menu_type=MenuType.Command, label='Open', accelerator='Ctrl-o',
                                 command=lambda *_: print("'File/Open' selected")))

    file_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    file_menu.add_child(MenuItem(name='save', menu_type=MenuType.Command, label='Save', accelerator='Ctrl-s',
                                 command=lambda *_: print("'File/Save' selected")))
    file_menu.add_child(MenuItem(name='save_as', menu_type=MenuType.Command, label='Save As',
                                 command=lambda *_: print("'File/Save As' selected")))

    file_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    file_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Exit', accelerator='Ctrl-e',
                                 command=lambda *_: print("'File/Exit' selected")))

    # -----------------------------------------------------------------------------------------
    # Print menu
    # -----------------------------------------------------------------------------------------
    print_menu = MenuItem(name='print', menu_type=MenuType.Cascade, label='Print')
    pdf_menu = MenuItem(menu_type=MenuType.Cascade, label='PDF')
    latex_menu = MenuItem(menu_type=MenuType.Cascade, label='Latex')
    print_menu.add_child(pdf_menu)
    print_menu.add_child(latex_menu)

    # pdf sub-menu
    pdf_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Teacher Schedules',
                                command=lambda *_: print("'Print/Teacher Schedules' selected")))
    pdf_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Lab Schedules',
                                command=lambda *_: print("'Print/Lab Schedules' selected")))
    pdf_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Stream Schedules',
                                command=lambda *_: print("'Print/Stream Schedules' selected")))

    pdf_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    pdf_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Text Output',
                                command=lambda *_: print("'Print/Text Output' selected")))

    # latex sub menu
    latex_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Teacher Schedules',
                                  command=lambda *_: print("'Print/Teacher Schedules' selected")))
    latex_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Lab Schedules',
                                  command=lambda *_: print("'Print/Lab Schedules' selected")))
    latex_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Stream Schedules',
                                  command=lambda *_: print("'Print/Stream Schedules' selected")))

    latex_menu.add_child(MenuItem(menu_type=MenuType.Separator))

    latex_menu.add_child(MenuItem(menu_type=MenuType.Command, label='Text Output',
                                  command=lambda *_: print("'Print/Text Output' selected")))

    # -----------------------------------------------------------------------------------------
    # _toolbar
    # -----------------------------------------------------------------------------------------
    toolbar_info = dict()
    toolbar_info['new'] = ToolbarItem(command=MenuItem.all_menu_items['new'].command, hint='Create new Schedule File')
    toolbar_info['open'] = ToolbarItem(command=MenuItem.all_menu_items['open'].command, hint='Open Schedule File')
    toolbar_info['save'] = ToolbarItem(command=MenuItem.all_menu_items['save'].command, hint='Save Schedule File')
    toolbar_order = ['new', 'open', '', 'save']

    # return list of top level menu items
    menu.append(file_menu)
    menu.append(print_menu)

    return toolbar_order, toolbar_info, menu


if __name__ == "__main__":
    main()



# =====================================================================================
# Notebook book-keeping
# =====================================================================================
from dataclasses import dataclass, field
from typing import Callable, Any


@dataclass
class NBTabInfo:
        name: str
        label: str
        subpages: list = field(default_factory=list)
        frame_args: dict[str, Any] =  field(default_factory=dict)
        creation_handler: Callable = lambda *_: None
        is_default_page: bool = False




import pytest

from src.scheduling_and_allocation.model import Stream


@pytest.fixture(scope="module", autouse=True)
def before_and_after_module():
    pass


@pytest.fixture(autouse=True)
def before_and_after():
    pass


# ============================================================================
# Stream
# ============================================================================
def test_descr_getter():
    descr = "Worst place in the world."
    stream = Stream("AB", descr)
    assert stream.description == descr


def test_number_getter():
    descr = "Worst place in the world."
    stream = Stream("AB", descr)
    assert stream.number == "AB"


def test_string_representation_short():
    descr = "Worst place in the world."
    stream = Stream("AB", descr)
    assert stream.number in str(stream)
    assert stream.description in str(stream)



from __future__ import annotations
from tkinter import *
import tkinter as tk
from typing import Any, Callable
from functools import partial
import os

from ..modified_tk.Scrolled import Scrolled
from ..modified_tk.FindImages import get_image_dir
from ..modified_tk.InitGuiFontsAndColours import TkColours

# TODO:  Maybe refactor the method for config_specs, etc.
#        Check DynamicTree methods (if was coded after this was)
# =================================================================================================
# NOTES TO DEVELOPERS
# =================================================================================================
# 1. columns and rows:
#    the 0th grid row is the label row, and the data row starts at 0, but is in grid row 1
#    the 0th grid column is the delete button, and the data column starts at 0, but is in grid column 1
#    if there is a button at the end of the row, the grid column will be the data column + 1
#    i.e.
#           grid_row = data_row + 1
#           grid_column = data_column + 1
#           del_button_column = 0
#           button_column = self.number_of_columns + 2
# -------------------------------------------------------------------------------------------------

class TableEntry(tk.Frame):
    """
    Creates a spreadsheet-ish widget, inside a frame

Inputs
------
    notebook: Frame
        the notebook widget that will contain the TableEntry widget

Additional TableEntry Options
-----------------------------
    rows: int
        The number of data rows in TableEntry, should not be modified after TableEntry is created
    columns: int
        The number of data columns in TableEntry, should not be modified after TableEntry is created
    titles: list
        A list of column titles
    colwidths: list
        A list of widths for each column (default = defwidth), should not be modified after TableEntry is created
    disabled: list
        A list containing a '1' for each column that is made 'disabled'.  Do not modify after TableEntry is created
    defwidth: int
        The default width of the columns (defaults to 8 if not defined). Do Not modify after TableEntry is created
    buttontext: str
        If buttoncmd is defined, then a button will be added to the row.  The text will be 'buttontext'.
        Do not modify after TableEntry is created
    delete: callable
        A function that will be called if the row is deleted.  The data in that row will be passed to the
        function as a list
    buttoncmd: callable
        If defined, a button will be created at the end of each row.  This function that will be called
        The data in that row will be passed to the function as a list

Public Methods
--------------

    add_empty_row()
        Adds a new row at the end of the DateEntry
    read_row(row:int)->List[Str]
        Reads the data in the row, and returns the data in a list (note that all data will be a string)
    put(row:int, col:int, datum:Any)
        Puts the string version of the data into the Entry widget at row,col
    cget(option:str)
        Returns the value of that particular option
    configure(option_name=value, ...)
        Sets the option. NB: Modifying some options will have no effect (see above)

Examples
--------

Example::

    data = [
        [1, "datum 1", "m1"],
        [2, "datum 2", "m2"],
        [3, "datum 3", "m3"],
    ]

    # make a Tk window
    mw = Tk()
    mw.geometry("500x300")
    frame = Frame(mw)
    frame.pack(expand=1, fill="both")

    # add the TableEntry object
    titles = ["one", "two", "three"]
    col_widths = [10, 20, 5]

    # note that buttoncmd must be here, it cannot be entered later via config
    de = TableEntry(frame, rows=0, columns=len(data[0]), colwidths=col_widths, titles=titles,buttoncmd=on_button_click)
    de.pack(side="top", expand=1, fill="both")
    de.configure(borderwidth=3)

    # disable the first column, but not the rest
    disabled = [1]
    de.configure(disabled=disabled)

    # callback for deleting a row
    de.configure(delete=on_delete)

    # populate with data
    for r in range(len(data)):
        for c in range(len(data[r])):
            de.put(r, c, data[r][c])

    de.add_empty_row();

    mw.mainloop()

    def on_delete(data: list):
        print(f"Deleted row, {data}")

    def on_button_click(data: list):
        print(f"Button clicked, {data}")
    """

    # ===================================================================
    # properties
    # ===================================================================

    # number of data rows
    @property
    def number_of_rows(self) -> int:
        """Gets and sets the number of rows in the table"""
        return self.cget("rows")

    # number of data columns
    @property
    def number_of_columns(self) -> int:
        """Gets and sets the number of columns in the table"""
        return self.cget("columns")

    # ===================================================================
    # Constructor
    # ===================================================================
    def __init__(self, parent: Frame, colours: TkColours, **kwargs):
        """Create a TableEntry widget inside 'notebook'"""
        tk.Frame.__init__(self, parent)
        self.colours = colours
        self._cget = dict()
        self.frame = None
        self.__scrolled_frame = None
        self.__delete_photo = None

        # ---------------------------------------------------------------
        # configuration and defaults
        # ---------------------------------------------------------------
        # this is a table of all the additional options available to
        # TableEntry, and the methods used to set those options
        self.__config_specs: dict[str:Callable[[TableEntry, Any], None]] = {
            'rows': self.__set_rows,
            'columns': self.__set_columns,
            'titles': self.__set_titles,
            'disabled': self.__set_disabled_columns,
            'delete': self.__set_delete_callback,
            'buttoncmd': self.__set_button_callback,
            'buttontext': self.__set_button_text,
            'colwidths': self.__set_column_widths,
            'defwidth': self.__set_default_width,
        }

        # this is a table of all the additional options available to
        # TableEntry, and their default values
        self._defaults: dict[str, Any] = {
            "rows": 10,
            "columns": 10,
            "titles": list(),
            "colwidths": list(),
            "disabled": list(),
            "defwidth": 8,
            "width": 100,
            "height": 200,
            "relief": 'flat',
            "buttontext": "Go",
            "delete": None,
            "buttoncmd": None,
        }

        # time_start the creation of the widget
        self.__populate_the_frame(**kwargs)

    # ====================================================================================
    # add empty_row
    # ====================================================================================
    def add_empty_row(self):
        """Add an empty row at the end of the 'spreadsheet'"""
        column_widths = self.cget("colwidths")
        columns_enabled_disabled = self.cget("disabled")
        if columns_enabled_disabled is None:
            columns_enabled_disabled = []

        # make sure lists are long enough
        while len(column_widths) < self.number_of_columns + 1:
            column_widths.append(0)
        while len(columns_enabled_disabled) < self.number_of_columns + 1:
            columns_enabled_disabled.append(False)

        # --------------------------------------------------------------------------------
        # add the grid_row
        # --------------------------------------------------------------------------------
        # always add new grid_row at the end of the existing EditResources widget
        grid_row = self.number_of_rows + 1

        # for each column, add an entry box
        for data_col in range(0, self.number_of_columns):

            # set the column width if defined, else use the default width
            column_widths[data_col] = column_widths[data_col] if column_widths[data_col] else self.cget("defwidth")

            # if there is something already there, delete it
            old = self.__get_widget_in_row_col(grid_row, data_col + 1)
            if old:
                old.destroy()

            # make entry widget
            w = Entry(self.frame,
                      width=column_widths[data_col],
                      bg=self.colours.DataBackground,
                      fg=self.colours.DataForeground,
                      disabledbackground=self.colours.DisabledBackground,
                      disabledforeground=self.colours.DataForeground,
                      relief="ridge",
                      )

            # if the column has been set to disabled, disable it
            if columns_enabled_disabled[data_col]:
                w.configure(state="disabled")

            # put widget in appropriate grid column and row
            w.grid(column=data_col + 1, sticky="nsew", row=grid_row)

            # key bindings for this entry widget
            w.bind("<Tab>", partial(self.__next_cell, w))
            w.bind("<Key-Return>", partial(self.__next_cell, w))
            w.bind("<Shift-Tab>", partial(self.__prev_cell, w))
            w.bind("<Key-Left>", partial(self.__prev_cell, w))
            w.bind("<Key-leftarrow>", partial(self.__prev_cell, w))
            w.bind("<Key-Up>", partial(self.__prev_row, w))
            w.bind("<Key-uparrow>", partial(self.__prev_row, w))
            w.bind("<Key-Down>", partial(self.__next_row, w))
            w.bind("<Key-downarrow>", partial(self.__next_row, w))
            w.bind("<Key-Right>", partial(self.__next_cell, w))
            w.bind("<Key-rightarrow>", partial(self.__next_cell, w))
            w.bind("<Button>", partial(self.__select_all, w))

            # I want my bindings to happen BEFORE the class bindings
            # TODO: This is not working properly, <TAB> still behaves weirdly
            bindtags = w.bindtags()
            w.bindtags((bindtags[1], bindtags[0], bindtags[2], bindtags[3]))

        # keep our grid_row count up to date
        self.configure(rows=grid_row)

        # add a `delete button` in the first column
        self.__add_delete_btn_to_row(grid_row)

        # if we need a grid_row button, add that to the last column
        self.__add_btn_to_end_of_row(grid_row)

        # we've added a new grid_row, so we need to adjust the size of the scrollable frame
        self.__adjust_size_of_frame()

    # ====================================================================================
    # read row
    # ====================================================================================
    def read_row(self, data_row: int) -> list[str]:
        """Get the data from the specified row"""
        data = list()
        grid_row = data_row + 1

        # cannot select the data from the title row
        if grid_row < 1:
            return data

        # get all the data from all the Entry widgets
        for data_col in range(0, self.number_of_columns):
            w = self.__get_widget_in_row_col(grid_row, data_col + 1)
            if w:
                value = w.get()
                value = value.strip()
                data.append(value)
        return data

    # ===================================================================
    # put data
    # ===================================================================
    def put(self, data_row: int, data_col: int, data: Any):
        """Put data into a cell specified by row and col"""
        if data_row < 0:
            return

        grid_row = data_row + 1
        grid_col = data_col + 1

        # add rows, if required
        if grid_row > self.number_of_rows:
            for _ in range(grid_row - self.number_of_rows):
                self.add_empty_row()

        # get the entry widget
        w = self.__get_widget_in_row_col(grid_row, grid_col)
        if not w or not isinstance(w, Entry):
            return

        # we have to disable Entry widgets before this can work
        disabled_flag = w.cget("state") == "disabled"
        if disabled_flag:
            w.configure(state="normal")

        # remove what was there and insert the data
        w.delete(0, 'end')
        w.insert(0, data)

        # if entry widget was disabled, set it back to disabled
        if disabled_flag:
            w.configure(state="disabled")

    # ===================================================================
    # clear data
    # ===================================================================
    def clear_data(self):
        """Put data into a cell specified by row and col"""
        for data_row in range(self.number_of_rows):
            grid_row = data_row + 1
            for data_col in range(self.number_of_columns):
                grid_col = data_col + 1

                # get the entry widget
                w = self.__get_widget_in_row_col(grid_row, grid_col)
                if not w or not isinstance(w, Entry):
                    return

                # we have to disable Entry widgets before this can work
                disabled_flag = w.cget("state") == "disabled"
                if disabled_flag:
                    w.configure(state="normal")

                # remove what was there and insert the data
                w.delete(0, 'end')

                # if entry widget was disabled, set it back to disabled
                if disabled_flag:
                    w.configure(state="disabled")

    # =================================================================================================
    # configure, cget, etc
    # =================================================================================================
    def cget(self, option: str) -> Any:
        if option in self._cget:
            return self._cget[option]
        else:
            return super().cget(option)

    def configure(self, **kwargs):
        for k, v in kwargs.items():
            if k in self._cget:
                self.__config_specs[k](v)
            else:
                super().configure(**{k: v})

    # ====================================================================================
    # add the user defined button at end of row
    # ====================================================================================
    def __add_btn_to_end_of_row(self, grid_row):
        if self.cget("buttoncmd"):
            b = Button(self.frame,
                       text=self.cget("buttontext"),
                       )
            b.configure(command=partial(self.__button_cmd, b))
            b.grid(column=self.number_of_columns + 1, sticky="nsew", row=grid_row)

    # ====================================================================================
    # add the "delete" button at the beginning of a row
    # ====================================================================================
    def __add_delete_btn_to_row(self, grid_row):
        b = Button(self.frame,
                   relief="flat",
                   highlightbackground=self.colours.HighlightBackground,
                   bg=self.colours.ButtonBackground,
                   )
        if self.delete_photo:
            b.configure(image=self.delete_photo)
        else:
            b.configure(text="del")
        b.configure(command=partial(self.__delete_row, b))
        b.grid(column=0, sticky="nsew", row=grid_row)

        # disable button?
        if self.__are_all_columns_disabled():
            b.configure(state='disabled')

    # ====================================================================================
    # after a row add or row delete, adjust the size of the frame
    # ====================================================================================
    def __adjust_size_of_frame(self):
        # --------------------------------------------------------------------------------
        # adjust height of pane
        # --------------------------------------------------------------------------------
        height = 0
        for row in range(self.number_of_rows + 1):
            w = self.__get_widget_in_row_col(row, 1)
            if not w:
                continue
            height += w.winfo_height()

        height += 2 * self.cget('border')
        height += self.__scrolled_frame.Subwidget("xscrollbar").winfo_height()
        self.frame.configure(height=height)
        self.update()

    # =================================================================================================
    # Generic function that applies calls a function for all Entry widgets
    # =================================================================================================
    def __apply_to_all_entry_widgets(self, func: callable):
        if self.number_of_columns and self.number_of_rows:
            for grid_col in range(1, self.number_of_columns + 1):
                for grid_row in range(1, self.number_of_rows + 1):
                    w = self.__get_widget_in_row_col(grid_row, grid_col)
                    if w is not None:
                        func(w, grid_row, grid_col)

    # ====================================================================================
    # are all columns disabled?
    # ====================================================================================
    def __are_all_columns_disabled(self) -> bool:
        disabled_columns = self.cget("disabled")
        if not disabled_columns:
            return False
        flag = True
        for col in disabled_columns:
            flag = flag and col
        return flag

    # ====================================================================================
    # callback for button at end of row being clicked
    # ====================================================================================
    def __button_cmd(self, btn, **kwargs):
        if self.cget("buttoncmd"):
            grid_row, grid_col = self.__get_row_col_of_widget(btn)
            data_row = grid_row - 1
            data = self.read_row(data_row)
            self.cget("buttoncmd")(data)

    # ====================================================================================
    # change focus to next cell (bound to Entry widgets)
    # ====================================================================================
    def __change_focus_to_new_cell(self, w: Entry, x_dir: int, y_dir: int, e:Event=None):

        if w:
            w.selection_clear()

        (grid_row, grid_col) = self.__get_row_col_of_widget(w)
        grid_col = grid_col + x_dir
        grid_row = grid_row + y_dir

        # if at end of row, scroll around to beginning of next row,
        # ignoring the 'delete' button
        if grid_col > self.number_of_columns:
            grid_row = grid_row + 1
            grid_col = 1

        # if at beginning of row, go to the end of the previous row
        if grid_col < 1:
            grid_row = grid_row - 1
            grid_col = self.number_of_columns

        # if we have exceeded the number of rows, create a new empty row
        if grid_row > self.number_of_rows:
            self.add_empty_row()
            #self.__scrolled_frame.update_scrollbars()

        # can't go less than zero
        if grid_row < 0:
            grid_row = 0

        # set the focus, and move the scrollbars appropriately
        w2 = self.__get_widget_in_row_col(grid_row, grid_col)
        if w2:
            self.__scrolled_frame.see(w2)
            self.update()
            if isinstance(w2, Entry):
                w2.focus_set()
                w2.selection_range(0, 'end')

                # if we land on a disabled widget, keep moving
                disabled_list = self.cget("disabled")
                if disabled_list[grid_col - 1]:
                    self.winfo_toplevel().after(20,lambda:self.__change_focus_to_new_cell(w2, x_dir, y_dir))

        self.update()

        # According to the documentation, this should prevent the event
        # from propagating to the next internal event handlers
        # (see bindtags), but it doesn't appear to work
        return "break"

    # ===================================================================
    # save the additional TableEntry options in a list
    # ===================================================================
    def __configure_save(self, **kwargs):
        for k, v in kwargs.items():
            self._cget[k] = v

    # ===================================================================
    # insert header_row
    # ===================================================================
    def __create_header_row(self):
        titles = self.cget("titles")
        col_widths = self.cget("colwidths")

        # make sure lists are long enough
        while len(titles) < self.number_of_columns + 1:
            titles.append("")

        while len(col_widths) < self.number_of_columns + 1:
            col_widths.append(0)

        # add labels to the top row
        for data_col in range(0, self.number_of_columns):
            col_widths[data_col] = col_widths[data_col] if col_widths[data_col] else self.cget("defwidth")
            w = Label(self.frame, text=titles[data_col], width=col_widths[data_col])
            w.grid(column=data_col + 1, sticky='nsew', row=0)

        self.__configure_save(colwidths=col_widths)

    # ====================================================================================
    # delete a row
    # ====================================================================================
    def __delete_row(self, b):
        grid_row, _ = self.__get_row_col_of_widget(b)
        if grid_row == 0:
            return  # cannot delete the header row
        data_row = grid_row - 1

        # run user supplied callback first
        delete_callback = self.cget("delete")
        if delete_callback:
            data = self.read_row(data_row)
            delete_callback(data)


        # we want to move everything up
        for grid_row in range(grid_row + 1, self.number_of_rows + 1):
            for grid_col in range(1, self.number_of_columns + 2):
                w1 = self.__get_widget_in_row_col(grid_row-1, grid_col)
                w2 = self.__get_widget_in_row_col(grid_row, grid_col)
                if w1:
                    was_state = w1.cget('state')
                    w1.configure(state="normal")
                    w1.delete(0, 'end')
                    if w2:
                        w1.insert(0, w2.get())
                    w1.configure(state=was_state)

        # adjust the number of rows
        self.configure(rows=self.number_of_rows - 1)

    # ===================================================================
    # draw the widgets onto the scrollable frame
    # ===================================================================
    def __draw(self):

        # create header row
        self.__create_header_row()

        # add rows
        num_of_rows = self.number_of_rows
        self.configure(rows=0)
        for r in range(0, num_of_rows):
            self.add_empty_row()

        # calculate the width of the row, to set the pane width
        width_total = 0
        for c in range(self.number_of_columns + 1):
            w = self.__get_widget_in_row_col(1, c)
            if w is not None:
                width_total += w.winfo_width() + 12

            if self.cget("buttoncmd"):
                width_total += 100

            width_total += 2 * self.cget("border")
            self.frame.configure(width=width_total)

    # ===================================================================
    # get row/col of a widget
    # ===================================================================
    def __get_row_col_of_widget(self, w: Entry):
        if w:
            i = w.grid_info()
            return i["row"], i["column"]
        return 0, 0

    # ===================================================================
    # get widget in row/col
    # ===================================================================
    def __get_widget_in_row_col(self, r, c) -> Entry | None:
        """get the widget in the row/col"""
        w = self.frame.grid_slaves(r, c)
        if w:
            return w[0]
        return None

    # ===================================================================
    def __next_cell(self, w: Any, *args, **kwargs):
        self.__change_focus_to_new_cell(w, 1, 0)

    # ===================================================================
    def __next_row(self, w: Any, *args, **kwargs):
        self.__change_focus_to_new_cell(w, 0, 1)

    # ===================================================================
    # populate the frame
    # ===================================================================
    def __populate_the_frame(self, **kwargs):

        # ---------------------------------------------------------------
        # create a scrolled pane
        # ---------------------------------------------------------------
        scrolled_frame = Scrolled(self, "Frame", scrollbars="se", border=2, relief="flat")
        scrolled_frame.pack(side='left', fill='both')
        self.__scrolled_frame = scrolled_frame
        self.frame = scrolled_frame.Subwidget("Frame")

        # ---------------------------------------------------------------
        # define the 'delete' image
        # ---------------------------------------------------------------
        image_dir = get_image_dir()
        self.delete_photo = PhotoImage(format='gif', file=os.path.join(image_dir, "ex.gif"))

        # ---------------------------------------------------------------
        # configure this table entry & draw
        # ---------------------------------------------------------------
        to_configure = self._defaults.copy()
        to_configure.update(kwargs)
        self.__configure_save(**to_configure)
        self.__draw()

    # ====================================================================================
    def __prev_cell(self, w: Any, *args, **kwargs):
        self.__change_focus_to_new_cell(w, -1, 0)

    # ====================================================================================
    def __prev_row(self, w: Any, *args, **kwargs):
        self.__change_focus_to_new_cell(w, 0, -1)

    # ====================================================================================
    # binding subroutine for clicking mouse button on entry widget
    # ====================================================================================
    def __select_all(self, w: Entry, *args, **kwargs):
        if w:
            w.focus()
            w.selection_clear()
            w.selection_range(0, 'end')

    # ====================================================================================
    # set the titles for each column
    # ====================================================================================
    def __set_titles(self, title_list: list):
        if title_list is None:
            return self.cget('titles')
        while len(title_list) < self.number_of_columns:
            title_list.append("")
        self.__configure_save(titles=title_list)
        self.__apply_to_all_entry_widgets(lambda w, r, c: w.configure(text=title_list[c - 1]))
        return self.cget('titles')

    # ====================================================================================
    # set which columns of entry widgets are disabled
    # ====================================================================================
    def __set_disabled_columns(self, column_disabled_list: list):
        if column_disabled_list is None:
            return self.cget("disabled")
        while len(column_disabled_list) < self.number_of_columns:
            column_disabled_list.append(0)
        self.__configure_save(disabled=column_disabled_list)

        def _disable_widget(w, r, c):
            if column_disabled_list[c - 1]:
                w.configure(state='disabled')
                w.configure(takefocus=0)
            else:
                w.configure(state='normal')

        self.__apply_to_all_entry_widgets(_disable_widget)
        return self.cget('disabled')

    # ====================================================================================
    # set the callback for the button at the end of the row
    # NB:  If this is not set during the __init__, there will NOT be a button at the end
    #      of the row
    # ====================================================================================
    def __set_button_callback(self, button_callback: callable):
        self.__configure_save(buttoncmd=button_callback)
        return self.cget("buttoncmd")

    # ====================================================================================
    # define the button's text (the button at the end of the row)
    # ====================================================================================
    def __set_button_text(self, button_text: str):
        self.__configure_save(buttontext=button_text)
        return self.cget("buttontext")

    # ====================================================================================
    # set the number of columns
    # ====================================================================================
    def __set_columns(self, columns: list):
        self.__configure_save(columns=columns)
        return self.cget("columns")

    # ====================================================================================
    # set the widths for each column
    # ====================================================================================
    def __set_column_widths(self, colwidths_list: list):
        if not colwidths_list:
            return self.cget("colwidths")
        self.__configure_save(colwidths=colwidths_list)

        def _colwidths(w, r, c):
            width = colwidths_list[c - 1] or self.cget("defwidth") or 8
            w.configure(width=width)

        self.__apply_to_all_entry_widgets(_colwidths)
        return self.cget('colwidths')

    # ====================================================================================
    # set the callback function for the delete button
    # ====================================================================================
    def __set_delete_callback(self, delete_callback: callable):
        self.__configure_save(delete=delete_callback)
        return self.cget("delete")

    # ====================================================================================
    # set the number of rows
    # ====================================================================================
    def __set_rows(self, rows: list):
        self.__configure_save(rows=rows)
        return self.cget("rows")

    # ====================================================================================
    # set the default width of a column
    # ====================================================================================
    def __set_default_width(self, default_width: int):
        self.__configure_save(defwidth=default_width)
        return self.cget("width")
